# Copyright (C) 2013  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import sqlite3
import pwd
import os
import platform
import grp


class SQLiteLogger(object):

    def _chown_db(self, path, uid_name='nobody', gid_name='nogroup'):
        path = path.rpartition("/")[0]
        if not os.path.isdir(path):
            os.mkdir(path)
        # TODO: Have this in a central place
        wanted_uid = pwd.getpwnam(uid_name)[2]
        # special handling for os x. (getgrname has trouble with gid below 0)
        if platform.mac_ver()[0]:
            wanted_gid = -2
        else:
            wanted_gid = grp.getgrnam(gid_name)[2]
        os.chown(path, wanted_uid, wanted_gid)

    def __init__(self, db_path="logs/conpot.db"):
        self._chown_db(db_path)
        self.conn = sqlite3.connect(db_path)
        self._create_db()

    def _create_db(self):
        cursor = self.conn.cursor()
        cursor.execute("""CREATE TABLE IF NOT EXISTS events
            (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                session TEXT,
                timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                remote TEXT,
                protocol TEXT,
                request TEXT,
                response TEXT
            )""")

    def log(self, event):
        cursor = self.conn.cursor()
        cursor.execute("INSERT INTO events(session, remote, protocol, request, response) VALUES (?, ?, ?, ?, ?)",
                       (str(event["id"]), str(event["remote"]), event['data_type'],
                        str(event["data"].get('request')), str(event["data"].get('response')))
        )
        self.conn.commit()
        return cursor.lastrowid

    def log_session(self, session):
        pass

    def select_data(self):
        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM events")
        print((cursor.fetchall()))



import sys
import logging
import random
from datetime import datetime

from pysnmp.entity.rfc3413 import cmdrsp
from pysnmp.proto import error
from pysnmp.proto.api import v2c
import pysnmp.smi.error
from pysnmp import debug
import gevent
import conpot.core as conpot_core

logger = logging.getLogger(__name__)


class conpot_extension(object):
    def _getStateInfo(self, snmpEngine, stateReference):
        for k, v in list(snmpEngine.messageProcessingSubsystems.items()):
            if stateReference in v._cache.__dict__['_Cache__stateReferenceIndex']:
                state_dict = v._cache.__dict__['_Cache__stateReferenceIndex'][stateReference][0]

        addr = state_dict['transportAddress']

        # msgVersion 0/1 to SNMPv1/2, msgversion 3 corresponds to SNMPv3
        if state_dict['msgVersion'] < 3:
            snmp_version = state_dict['msgVersion'] + 1
        else:
            snmp_version = state_dict['msgVersion']

        return addr, snmp_version

    def log(self, version, msg_type, addr, req_varBinds, res_varBinds=None):
        session = conpot_core.get_session('snmp', addr[0], addr[1])
        req_oid = req_varBinds[0][0]
        req_val = req_varBinds[0][1]
        event_type = 'SNMPv{0} {1}'.format(version, msg_type)
        request = {'oid': str(req_oid), 'val': str(req_val)}
        response = None

        logger.info('%s request from %s: %s %s', event_type, addr, req_oid, req_val)

        if res_varBinds:
            res_oid = ".".join(map(str, res_varBinds[0][0]))
            res_val = res_varBinds[0][1]
            logger.info('%s response to %s: %s %s', event_type, addr, res_oid, res_val)
            response = {'oid': str(res_oid), 'val': str(res_val)}

        session.add_event({'type': event_type, 'request': request, 'response': response})

    def do_tarpit(self, delay):

        # sleeps the thread for $delay ( should be either 1 float to apply a static period of time to sleep,
        # or 2 floats seperated by semicolon to sleep a randomized period of time determined by ( rand[x;y] )

        lbound, _, ubound = delay.partition(";")

        if not lbound or lbound is None:
            # no lower boundary found. Assume zero latency
            pass
        elif not ubound or ubound is None:
            # no upper boundary found. Assume static latency
            gevent.sleep(float(lbound))
        else:
            # both boundaries found. Assume random latency between lbound and ubound
            gevent.sleep(random.uniform(float(lbound), float(ubound)))

    def check_evasive(self, state, threshold, addr, cmd):

        # checks if current states are > thresholds and returns True if the request
        # is considered to be a DoS request.

        state_individual, state_overall = state
        threshold_individual, _, threshold_overall = threshold.partition(';')

        if int(threshold_individual) > 0:
            if int(state_individual) > int(threshold_individual):
                logger.warning('SNMPv%s: DoS threshold for %s exceeded (%s/%s).', cmd, addr, state_individual,
                               threshold_individual)
                # DoS threshold exceeded.
                return True

        if int(threshold_overall) > 0:
            if int(state_overall) > int(threshold_overall):
                logger.warning('SNMPv%s: DDoS threshold exceeded (%s/%s).', cmd, state_individual, threshold_overall)
                # DDoS threshold exceeded
                return True

        # This request will be answered
        return False


class c_GetCommandResponder(cmdrsp.GetCommandResponder, conpot_extension):
    def __init__(self, snmpEngine, snmpContext, databus_mediator):
        self.databus_mediator = databus_mediator
        self.tarpit = '0;0'
        self.threshold = '0;0'

        cmdrsp.GetCommandResponder.__init__(self, snmpEngine, snmpContext)
        conpot_extension.__init__(self)

    def handleMgmtOperation(
            self, snmpEngine, stateReference, contextName, PDU, acInfo):
        (acFun, acCtx) = acInfo
        # rfc1905: 4.2.1.1
        mgmtFun = self.snmpContext.getMibInstrum(contextName).readVars

        varBinds = v2c.apiPDU.getVarBinds(PDU)
        addr, snmp_version = self._getStateInfo(snmpEngine, stateReference)

        evasion_state = self.databus_mediator.update_evasion_table(addr)
        if self.check_evasive(evasion_state, self.threshold, addr, str(snmp_version)+' Get'):
            return None

        rspVarBinds = None
        try:
            # generate response
            rspVarBinds = mgmtFun(v2c.apiPDU.getVarBinds(PDU), (acFun, acCtx))

            # determine the correct response class and update the dynamic value table
            reference_class = rspVarBinds[0][1].__class__.__name__
            reference_value = rspVarBinds[0][1]

            response = self.databus_mediator.get_response(reference_class, tuple(rspVarBinds[0][0]))
            if response:
                rspModBinds = [(tuple(rspVarBinds[0][0]), response)]
                rspVarBinds = rspModBinds

        finally:
            self.log(snmp_version, 'Get', addr, varBinds, rspVarBinds)

        # apply tarpit delay
        if self.tarpit is not 0:
            self.do_tarpit(self.tarpit)

        # send response
        self.sendRsp(snmpEngine, stateReference, 0, 0, rspVarBinds)
        self.releaseStateInformation(stateReference)


class c_NextCommandResponder(cmdrsp.NextCommandResponder, conpot_extension):
    def __init__(self, snmpEngine, snmpContext, databus_mediator):
        self.databus_mediator = databus_mediator
        self.tarpit = '0;0'
        self.threshold = '0;0'

        cmdrsp.NextCommandResponder.__init__(self, snmpEngine, snmpContext)
        conpot_extension.__init__(self)

    def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
        (acFun, acCtx) = acInfo
        # rfc1905: 4.2.2.1

        mgmtFun = self.snmpContext.getMibInstrum(contextName).readNextVars
        varBinds = v2c.apiPDU.getVarBinds(PDU)

        addr, snmp_version = self._getStateInfo(snmpEngine, stateReference)

        evasion_state = self.databus_mediator.update_evasion_table(addr)
        if self.check_evasive(evasion_state, self.threshold, addr, str(snmp_version)+' GetNext'):
            return None

        rspVarBinds = None
        try:
            while 1:
                rspVarBinds = mgmtFun(varBinds, (acFun, acCtx))

                # determine the correct response class and update the dynamic value table
                reference_class = rspVarBinds[0][1].__class__.__name__
                reference_value = rspVarBinds[0][1]

                response = self.databus_mediator.get_response(reference_class, tuple(rspVarBinds[0][0]))
                if response:
                    rspModBinds = [(tuple(rspVarBinds[0][0]), response)]
                    rspVarBinds = rspModBinds

                # apply tarpit delay
                if self.tarpit is not 0:
                    self.do_tarpit(self.tarpit)

                # send response
                try:
                    self.sendRsp(snmpEngine, stateReference, 0, 0, rspVarBinds)
                except error.StatusInformation:
                    idx = sys.exc_info()[1]['idx']
                    varBinds[idx] = (rspVarBinds[idx][0], varBinds[idx][1])
                else:
                    break

        finally:
            self.log(snmp_version, 'GetNext', addr, varBinds, rspVarBinds)

        self.releaseStateInformation(stateReference)


class c_BulkCommandResponder(cmdrsp.BulkCommandResponder, conpot_extension):
    def __init__(self, snmpEngine, snmpContext, databus_mediator):
        self.databus_mediator = databus_mediator
        self.tarpit = '0;0'
        self.threshold = '0;0'

        cmdrsp.BulkCommandResponder.__init__(self, snmpEngine, snmpContext)
        conpot_extension.__init__(self)

    def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
        (acFun, acCtx) = acInfo
        nonRepeaters = v2c.apiBulkPDU.getNonRepeaters(PDU)
        if nonRepeaters < 0:
            nonRepeaters = 0
        maxRepetitions = v2c.apiBulkPDU.getMaxRepetitions(PDU)
        if maxRepetitions < 0:
            maxRepetitions = 0

        reqVarBinds = v2c.apiPDU.getVarBinds(PDU)
        addr, snmp_version = self._getStateInfo(snmpEngine, stateReference)

        evasion_state = self.databus_mediator.update_evasion_table(addr)
        if self.check_evasive(evasion_state, self.threshold, addr, str(snmp_version)+' Bulk'):
            return None
        raise Exception('This class is not converted to new architecture')
        try:
            N = min(int(nonRepeaters), len(reqVarBinds))
            M = int(maxRepetitions)
            R = max(len(reqVarBinds) - N, 0)

            if R: M = min(M, self.maxVarBinds / R)

            debug.logger & debug.flagApp and debug.logger('handleMgmtOperation: N %d, M %d, R %d' % (N, M, R))

            mgmtFun = self.snmpContext.getMibInstrum(contextName).readNextVars

            if N:
                rspVarBinds = mgmtFun(reqVarBinds[:N], (acFun, acCtx))
            else:
                rspVarBinds = []

            varBinds = reqVarBinds[-R:]
            while M and R:
                rspVarBinds.extend(
                    mgmtFun(varBinds, (acFun, acCtx))
                )
                varBinds = rspVarBinds[-R:]
                M = M - 1
        finally:
            self.log(snmp_version, 'Bulk', addr, varBinds, rspVarBinds)

        # apply tarpit delay
        if self.tarpit is not 0:
            self.do_tarpit(self.tarpit)

        # send response
        if len(rspVarBinds):
            self.sendRsp(snmpEngine, stateReference, 0, 0, rspVarBinds)
            self.releaseStateInformation(stateReference)
        else:
            raise pysnmp.smi.error.SmiError()


class c_SetCommandResponder(cmdrsp.SetCommandResponder, conpot_extension):
    def __init__(self, snmpEngine, snmpContext, databus_mediator):
        self.databus_mediator = databus_mediator
        self.tarpit = '0;0'
        self.threshold = '0;0'

        conpot_extension.__init__(self)
        cmdrsp.SetCommandResponder.__init__(self, snmpEngine, snmpContext)

    def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
        (acFun, acCtx) = acInfo

        mgmtFun = self.snmpContext.getMibInstrum(contextName).writeVars

        varBinds = v2c.apiPDU.getVarBinds(PDU)
        addr, snmp_version = self._getStateInfo(snmpEngine, stateReference)

        evasion_state = self.databus_mediator.update_evasion_table(addr)
        if self.check_evasive(evasion_state, self.threshold, addr, str(snmp_version)+' Set'):
            return None

        # rfc1905: 4.2.5.1-13
        rspVarBinds = None

        # apply tarpit delay
        if self.tarpit is not 0:
            self.do_tarpit(self.tarpit)

        try:
            rspVarBinds = mgmtFun(v2c.apiPDU.getVarBinds(PDU), (acFun, acCtx))

            # generate response
            self.sendRsp(snmpEngine, stateReference, 0, 0, rspVarBinds)
            self.releaseStateInformation(stateReference)

            oid = tuple(rspVarBinds[0][0])
            self.databus_mediator.set_value(oid, rspVarBinds[0][1])

        except (pysnmp.smi.error.NoSuchObjectError,
                pysnmp.smi.error.NoSuchInstanceError):
            e = pysnmp.smi.error.NotWritableError()
            e.update(sys.exc_info()[1])
            raise e
        finally:
            self.log(snmp_version, 'Set', addr, varBinds, rspVarBinds)



# Copyright (C) 2013 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gevent import subprocess
import logging
import os
import re

from pysmi.reader.localfile import FileReader
from pysmi.reader.httpclient import HttpReader
from pysmi.searcher.pyfile import PyFileSearcher
from pysmi.searcher.pypackage import PyPackageSearcher
from pysmi.searcher.stub import StubSearcher
from pysmi.writer.pyfile import PyFileWriter
from pysmi.parser.smi import SmiV2Parser
from pysmi.codegen.pysnmp import PySnmpCodeGen, baseMibs
from pysmi.compiler import MibCompiler

logger = logging.getLogger(__name__)

BUILD_SCRIPT = 'build-pysnmp-mib'

# dict of lists, where the list contain the dependency names for the given dict key
mib_dependency_map = {}
compiled_mibs = []
# key = mib name, value = full path to the file
file_map = {}


def mib2pysnmp(mib_file, output_dir):
    """
    The 'build-pysnmp-mib' script we previously used is no longer available
    Latest pysmi has the ability to generate a .py file from .mib automatically

    :param mib_file: path to the .mib file we want to compile
    :param output_dir: path to the output directory
    :return: True if we successfully compile the .mib to a .py
    """

    logger.debug('Compiling mib file: %s', mib_file)

    # create a mib compiler with output dir
    mibCompiler = MibCompiler(SmiV2Parser(), PySnmpCodeGen(),
                              PyFileWriter(output_dir))

    # add sources from where we fetch dependencies
    mibCompiler.addSources(HttpReader('mibs.snmplabs.com', 80, '/asn1/@mib@'))
    mibCompiler.addSources(
        FileReader(os.path.dirname(os.path.abspath(mib_file))))

    # add searchers
    mibCompiler.addSearchers(PyFileSearcher(output_dir))
    mibCompiler.addSearchers(PyPackageSearcher('pysnmp.mibs'))
    mibCompiler.addSearchers(StubSearcher(*baseMibs))

    # compile, there should be a MIBFILE.py generated under output_dir
    mibName = os.path.basename(mib_file).replace('.mib', '')
    results = mibCompiler.compile(mibName)

    if results[mibName] == 'compiled' or results[mibName] == 'untouched':
        return True

    return False


def _get_files(raw_mibs_dir, recursive):
    for dir_path, dirs, files in os.walk(raw_mibs_dir, followlinks=True):
        for file_name in files:
            yield os.path.join(dir_path, file_name)
        if not recursive:
            break


def generate_dependencies(data, mib_name):
    """
    Parses a MIB for dependencies and populates an internal dependency map.
    :param data: A string representing an entire MIB file (string).
    :param mib_name: Name of the MIB (string).
    """
    if mib_name not in mib_dependency_map:
        mib_dependency_map[mib_name] = []
    imports_section_search = re.search('IMPORTS(?P<imports_section>.*?);', data, re.DOTALL)
    if imports_section_search:
        imports_section = imports_section_search.group('imports_section')
        for dependency in re.finditer('FROM (?P<mib_name>[\w-]+)', imports_section):
            dependency_name = dependency.group('mib_name')
            if dependency_name not in mib_dependency_map:
                mib_dependency_map[dependency_name] = []
            mib_dependency_map[mib_name].append(dependency_name)


def find_mibs(raw_mibs_dirs, recursive=True):
    """
    Scans for MIB files and populates an internal MIB->path mapping.
    :param raw_mibs_dirs: Directories to search for MIB files (list of strings).
    :param recursive:  If True raw_mibs_dirs will be scanned recursively.
    :return: A list of found MIB names (list of strings).
    """
    files_scanned = 0
    for raw_mibs_dir in raw_mibs_dirs:
        for _file in _get_files(raw_mibs_dir, recursive):
            files_scanned += 1
            # making sure we don't start parsing some epic file
            if os.path.getsize(_file) > int('1048576'):
                continue
            with open(_file) as _mibfile:
                data = _mibfile.read()
                # 2048 - just like a rock star.
                mib_search = re.search('(?P<mib_name>[\w-]+) DEFINITIONS ::= BEGIN', data[0:2048], re.IGNORECASE)
                if mib_search:
                    mib_name = mib_search.group('mib_name')
                    file_map[mib_name] = _file
                    generate_dependencies(data, mib_name)
        logging.debug('Done scanning for mib files, recursive scan was initiated from {0} directories and found {1} '
                      'MIB files of {2} scanned files.'
                      .format(len(raw_mibs_dirs), len(file_map), files_scanned))
    return list(file_map.keys())


def compile_mib(mib_name, output_dir):
    """
    Compiles the given mib_name if it is found in the internal MIB file map. If the MIB depends on other MIBs,
    these will get compiled automatically.
    :param mib_name: Name of mib to compile (string).
    :param output_dir: Output directory (string).
    """
    # resolve dependencies recursively

    for dependency in mib_dependency_map[mib_name]:
        if dependency not in compiled_mibs and dependency in file_map:
            compile_mib(dependency, output_dir)

    result = mib2pysnmp(file_map[mib_name], output_dir)
    if result:
        compiled_mibs.append(mib_name)



# modified by Sooky Peter <xsooky00@stud.fit.vutbr.cz>
# Brno University of Technology, Faculty of Information Technology
import struct
from lxml import etree
import codecs
from modbus_tk.modbus import Databank, DuplicatedKeyError, MissingKeyError, \
                             ModbusInvalidRequestError
from modbus_tk import defines

from conpot.protocols.modbus.slave import MBSlave
import logging

logger = logging.getLogger(__name__)


class SlaveBase(Databank):

    """
    Database keeping track of the slaves.
    """

    def __init__(self, template):
        Databank.__init__(self)
        self.dom = etree.parse(template)

    def add_slave(self, slave_id, unsigned=True, memory=None):
        """
        Add a new slave with the given id
        """
        if (slave_id < 0) or (slave_id > 255):
            raise Exception("Invalid slave id %d" % slave_id)
        if slave_id not in self._slaves:
            self._slaves[slave_id] = MBSlave(slave_id, self.dom)
            return self._slaves[slave_id]
        else:
            raise DuplicatedKeyError("Slave %d already exists" % slave_id)

    def handle_request(self, query, request, mode):
        """
        Handles a request. Return value is a tuple where element 0
        is the response object and element 1 is a dictionary
        of items to log.
        """
        request_pdu = None
        response_pdu = b''
        slave_id = None
        function_code = None
        func_code = None
        slave = None
        response = None

        try:
            # extract the pdu and the slave id
            slave_id, request_pdu = query.parse_request(request)
            if len(request_pdu) > 0:
                (func_code, ) = struct.unpack(">B", request_pdu[:1])

            logger.debug("Working mode: %s" % mode)

            if mode == 'tcp':
                if slave_id == 0 or slave_id == 255:
                    slave = self.get_slave(slave_id)
                    response_pdu = slave.handle_request(request_pdu)
                    response = query.build_response(response_pdu)
                else:
                    # TODO:
                    # Shall we return SLAVE DEVICE FAILURE, or ILLEGAL ACCESS?
                    # Would it be better to make this configurable?
                    r = struct.pack(
                        ">BB", func_code + 0x80, defines.SLAVE_DEVICE_FAILURE)
                    response = query.build_response(r)

            elif mode == 'serial':
                if slave_id == 0:           # broadcasting
                    for key in self._slaves:
                        response_pdu = self._slaves[key].handle_request(
                            request_pdu, broadcast=True)

                    # no response is sent back
                    return (None, {'request': request_pdu.encode('hex'),
                                   'slave_id': slave_id,
                                   'function_code': func_code,
                                   'response': ''})
                elif 0 < slave_id <= 247:   # normal request handling
                    slave = self.get_slave(slave_id)
                    response_pdu = slave.handle_request(request_pdu)
                    # make the full response
                    response = query.build_response(response_pdu)
                else:
                    # TODO:
                    # Same here. Return SLAVE DEVICE FAILURE or ILLEGAL ACCESS?
                    r = struct.pack(
                        ">BB", func_code + 0x80, defines.SLAVE_DEVICE_FAILURE)
                    response = query.build_response(r)

        except (MissingKeyError, IOError) as e:
            logger.error(e)
            # If slave was not found or the request was not handled correctly,
            # return a server error response
            r = struct.pack(
                ">BB", func_code + 0x80, defines.SLAVE_DEVICE_FAILURE)
            response = query.build_response(r)
        except ModbusInvalidRequestError as e:
            logger.error(e)
            # TODO: return something here?

        if slave:
            function_code = slave.function_code

        return (response, {'request': codecs.encode(request_pdu, 'hex'),
                           'slave_id': slave_id,
                           'function_code': function_code,
                           'response': codecs.encode(response_pdu, 'hex')})




import struct
import logging

from modbus_tk.modbus import Slave, ModbusError, ModbusInvalidRequestError, InvalidArgumentError, DuplicatedKeyError,\
                             InvalidModbusBlockError, OverlapModbusBlockError
from modbus_tk import defines, utils
from conpot.helpers import str_to_bytes
from .modbus_block_databus_mediator import ModbusBlockDatabusMediator

logger = logging.getLogger(__name__)


class MBSlave(Slave):

    """
    Customized Modbus slave representation extending modbus_tk.modbus.Slave
    """

    def __init__(self, slave_id, dom):
        Slave.__init__(self, slave_id)
        self._fn_code_map = {defines.READ_COILS: self._read_coils,
                             defines.READ_DISCRETE_INPUTS: self._read_discrete_inputs,
                             defines.READ_INPUT_REGISTERS: self._read_input_registers,
                             defines.READ_HOLDING_REGISTERS: self._read_holding_registers,
                             defines.WRITE_SINGLE_COIL: self._write_single_coil,
                             defines.WRITE_SINGLE_REGISTER: self._write_single_register,
                             defines.WRITE_MULTIPLE_COILS: self._write_multiple_coils,
                             defines.WRITE_MULTIPLE_REGISTERS: self._write_multiple_registers,
                             defines.DEVICE_INFO: self._device_info,
                             defines.REPORT_SLAVE_ID: self._report_slave_id,
                             }
        self.dom = dom
        logger.debug("Modbus slave (ID: %d) created" % self._id)

    def _report_slave_id(self, request_pdu):
        logger.debug('Requested to report slave ID (0x11)')
        response = struct.pack(">B", 0x11)  # function code
        response += struct.pack(">B", 1)    # byte count
        response += struct.pack(">B", 1)    # slave id
        response += struct.pack(">B", 0xFF) # run status, OxFF on, 0x00 off
        return response

    def _device_info(self, request_pdu):
        info_root = self.dom.xpath('//modbus/device_info')[0]
        vendor_name = info_root.xpath('./VendorName/text()')[0]
        product_code = info_root.xpath('./ProductCode/text()')[0]
        major_minor_revision = info_root.xpath('./MajorMinorRevision/text()')[0]

        (req_device_id, req_object_id) = struct.unpack(">BB", request_pdu[2:4])
        device_info = {
            0: vendor_name,
            1: product_code,
            2: major_minor_revision
        }

        # MEI type
        response = struct.pack(">B", 0x0E)
        # requested device id
        response += struct.pack(">B", req_device_id)
        # conformity level
        response += struct.pack(">B", 0x01)
        # followup data 0x00 is False
        response += struct.pack(">B", 0x00)
        # No next object id
        response += struct.pack(">B", 0x00)
        # Number of objects
        response += struct.pack(">B", len(device_info))
        for i in range(len(device_info)):
            # Object id
            response += struct.pack(">B", i)
            # Object length
            response += struct.pack(">B", len(device_info[i]))
            response += str_to_bytes(device_info[i])
        return response

    def handle_request(self, request_pdu, broadcast=False):
        """
        parse the request pdu, makes the corresponding action
        and returns the response pdu
        """

        logger.debug("Slave (ID: %d) is handling request" % self._id)

        with self._data_lock:  # thread-safe
            try:
                # get the function code
                (self.function_code, ) = struct.unpack(">B", request_pdu[:1])

                # check if the function code is valid. If not returns error response
                if not self.function_code in self._fn_code_map:
                    raise ModbusError(defines.ILLEGAL_FUNCTION)

                can_broadcast = [defines.WRITE_MULTIPLE_COILS, defines.WRITE_MULTIPLE_REGISTERS,
                                 defines.WRITE_SINGLE_COIL, defines.WRITE_SINGLE_REGISTER]
                if broadcast and (self.function_code not in can_broadcast):
                    raise ModbusInvalidRequestError("Function %d can not be broadcasted" % self.function_code)

                # execute the corresponding function
                try:
                    response_pdu = self._fn_code_map[self.function_code](request_pdu)
                except struct.error:
                    raise ModbusError(exception_code=3)
                if response_pdu:
                    if broadcast:
                        # not really sure whats going on here - better log it!
                        logger.info("Modbus broadcast: %s" % (utils.get_log_buffer("!!", response_pdu)))
                        return ""
                    else:
                        return struct.pack(">B", self.function_code) + response_pdu
                raise Exception("No response for function %d" % self.function_code)

            except ModbusError as e:
                logger.error('Exception caught: %s. (A proper response will be sent to the peer)', e)
                return struct.pack(">BB", self.function_code + 128, e.get_exception_code())

    def add_block(self, block_name, block_type, starting_address, size):
        """Add a new block identified by its name"""
        with self._data_lock:  # thread-safe
            if size <= 0:
                raise InvalidArgumentError("size must be a positive number")
            if starting_address < 0:
                raise InvalidArgumentError("starting address must be zero or positive number")
            if block_name in self._blocks:
                raise DuplicatedKeyError("Block %s already exists. " % block_name)

            if block_type not in self._memory:
                raise InvalidModbusBlockError("Invalid block type %d" % block_type)

            # check that the new block doesn't overlap an existing block
            # it means that only 1 block per type must correspond to a given address
            # for example: it must not have 2 holding registers at address 100
            index = 0
            for i in range(len(self._memory[block_type])):
                block = self._memory[block_type][i]
                if block.is_in(starting_address, size):
                    raise OverlapModbusBlockError("Overlap block at %d size %d" % (block.starting_address, block.size))
                if block.starting_address > starting_address:
                    index = i
                    break

            # if the block is ok: register it
            self._blocks[block_name] = (block_type, starting_address)
            # add it in the 'per type' shortcut
            self._memory[block_type].insert(index, ModbusBlockDatabusMediator(block_name, starting_address))



# Copyright (C) 2013  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
import gevent

from conpot.core.loggers.hpfriends import HPFriendsLogger


class Test_HPFriends(unittest.TestCase):
    @unittest.skip('disabled until honeycloud up and running again')
    def test_hpfriends(self):
        """
        Objective: Test if data can be published to hpfriends without errors.
        """

        host = 'hpfriends.honeycloud.net'
        port = 20000
        ident = 'HBmU08rR'
        secret = 'XDNNuMGYUuWFaWyi'
        channels = ["test.test", ]
        hpf = HPFriendsLogger(host, port, ident, secret, channels)
        gevent.sleep(0.5)
        error_message = hpf.log('some some test data')
        gevent.sleep(0.5)
        self.assertIsNone(error_message, 'Unexpected error message: {0}'.format(error_message))



# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import socket

import gevent
from gevent.server import StreamServer
import conpot
import conpot.core as conpot_core
from conpot.protocols.kamstrup.management_protocol.command_responder import CommandResponder
from conpot.helpers import str_to_bytes
from conpot.core.protocol_wrapper import conpot_protocol

logger = logging.getLogger(__name__)


@conpot_protocol
class KamstrupManagementServer(object):
    def __init__(self, template, template_directory, args, timeout=0):
        self.template = template
        self.timeout = timeout
        self.command_responder = CommandResponder()
        self.banner = "\r\nWelcome...\r\nConnected to [{0}]\r\n"
        logger.info('Kamstrup management protocol server initialized.')
        self.server = None

    def handle(self, sock, address):
        session = conpot_core.get_session('kamstrup_management_protocol', address[0], address[1])
        logger.info('New Kamstrup connection from %s:%s. (%s)', address[0], address[1], session.id)
        session.add_event({'type': 'NEW_CONNECTION'})

        try:
            sock.send(
                str_to_bytes(self.banner.format(conpot_core.get_databus().get_value("mac_address")))
            )

            while True:
                data = sock.recv(1024)
                if not data:
                    logger.info('Kamstrup client disconnected. (%s)', session.id)
                    session.add_event({'type': 'CONNECTION_LOST'})
                    break
                request = data.decode()
                logdata = {'request': request}
                response = self.command_responder.respond(request)
                logdata['response'] = response
                logger.info('Kamstrup management traffic from %s: %s (%s)', address[0], logdata, session.id)
                session.add_event(logdata)
                gevent.sleep(0.25)  # TODO measure delay and/or RTT

                if response is None:
                    session.add_event({'type': 'CONNECTION_LOST'})
                    break
                # encode data before sending
                reply = str_to_bytes(response)
                sock.send(reply)

        except socket.timeout:
            logger.debug('Socket timeout, remote: %s. (%s)', address[0], session.id)
            session.add_event({'type': 'CONNECTION_LOST'})

        sock.close()

    def start(self, host, port):
        connection = (host, port)
        self.server = StreamServer(connection, self.handle)
        logger.info('Kamstrup management protocol server started on: %s', connection)
        self.server.serve_forever()

    def stop(self):
        self.server.stop()


if __name__ == '__main__':
    TCP_IP = '127.0.0.1'
    TCP_PORT = 50100
    import os
    dir_name = os.path.dirname(conpot.__file__)
    conpot_core.get_databus().initialize(dir_name + '/templates/kamstrup_382/template.xml')
    server = KamstrupManagementServer(dir_name + '/templates/kamstrup_382/kamstrup_management/kamstrup_management.xml',
                                      None, None)
    try:
        server.start(TCP_IP, TCP_PORT)
    except KeyboardInterrupt:
        server.stop()


# Copyright (C) 2017  Yuru Shao <shaoyuru@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent.monkey; gevent.monkey.patch_all()
import unittest
import os
from gevent import socket
import conpot
from conpot.protocols.enip.enip_server import EnipServer
from cpppo.server.enip import client


class TestENIPServer(unittest.TestCase):
    def setUp(self):
        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)
        template = self.dir_name + '/templates/default/enip/enip.xml'
        # start the tcp server
        self.enip_server_tcp= EnipServer(template, None, None)
        self.enip_server_tcp.port = 50002
        self.server_greenlet_tcp = gevent.spawn(self.enip_server_tcp.start, self.enip_server_tcp.addr,
                                                self.enip_server_tcp.port)
        self.server_greenlet_tcp.start()

        # start the udp server
        self.enip_server_udp = EnipServer(template, None, None)
        self.enip_server_udp.config.mode = 'udp'
        self.enip_server_udp.port = 60002
        self.server_greenlet_udp = gevent.spawn(self.enip_server_udp.start, self.enip_server_udp.addr,
                                                self.enip_server_udp.port)
        self.server_greenlet_udp.start()

    def tearDown(self):
        self.enip_server_tcp.stop()
        self.enip_server_udp.stop()

    def attribute_operations(self, paths, int_type=None, **kwds):
        for op in client.parse_operations(paths, int_type=int_type or 'SINT', **kwds):
            path_end = op['path'][-1]
            if 'instance' in path_end:
                op['method'] = 'get_attributes_all'
                assert 'data' not in op, "All Attributes cannot be operated on using Set Attribute services"
            elif 'symbolic' in path_end or 'attribute' in path_end or 'element':
                op['method'] = 'set_attribute_single' if 'data' in op else 'get_attribute_single'
            else:
                raise AssertionError("Path invalid for Attribute services: %r", op['path'])
            yield op

    def test_read_tags(self):
        with client.connector(host=self.enip_server_tcp.addr,
                              port=self.enip_server_tcp.port, timeout=4.0) as connection:
            tags = ['@22/1/1']
            ops = self.attribute_operations(tags)
            for idx, dsc, op, rpy, sts, val in connection.pipeline(operations=ops):
                self.assertEqual(100, val[0])

    def test_write_tags(self):
        with client.connector(host=self.enip_server_tcp.addr,
                              port=self.enip_server_tcp.port, timeout=4.0) as connection:
            tags = ['@22/1/1=(SINT)50', '@22/1/1']
            ops = self.attribute_operations(tags)
            for idx, dsc, op, rpy, sts, val in connection.pipeline(operations=ops):
                if idx == 0:
                    self.assertEqual(True, val)
                elif idx == 1:
                    self.assertEqual(50, val[0])

    # TCP Tests
    def test_list_services_tcp(self):
        # test tcp
        with client.connector(host=self.enip_server_tcp.addr,
                              port=self.enip_server_tcp.port, timeout=4.0,
                              udp=False, broadcast=False) as connection:
            connection.list_services()
            connection.shutdown()
            while True:
                response, ela = client.await(connection, timeout=4.0)
                if response:
                    self.assertEqual("Communications",
                                     response['enip']['CIP']['list_services']['CPF']['item'][0]['communications_service']['service_name'])
                else:
                    break

    def test_list_services_udp(self):
        # test udp
        with client.connector(host=self.enip_server_udp.addr,
                              port=self.enip_server_udp.port, timeout=4.0,
                              udp=True, broadcast=True) as connection:
            connection.list_services()
            # TODO: udp does not cleanly shutdown. We get OSError.
            while True:
                response, ela = client.await(connection, timeout=4.0)
                if response:
                    self.assertEqual("Communications",
                                     response['enip']['CIP']['list_services']['CPF']['item'][0]['communications_service']['service_name'])
                else:
                    break

    def test_list_identity_tcp(self):
        # test tcp
        with client.connector(host=self.enip_server_tcp.addr,
                              port=self.enip_server_tcp.port, timeout=4.0,
                              udp=False, broadcast=False) as connection:
            connection.list_identity()
            connection.shutdown()
            while True:
                response, ela = client.await(connection, timeout=4.0)
                if response:
                    expected = self.enip_server_tcp.config.product_name
                    self.assertEqual(expected, response['enip']['CIP']['list_identity']['CPF']['item'][0]['identity_object']['product_name'])
                else:
                    break

    def test_list_identity_udp(self):
        with client.connector(host=self.enip_server_udp.addr,
                              port=self.enip_server_udp.port, timeout=4.0,
                              udp=True, broadcast=True) as connection:
            connection.list_identity()
            # TODO: udp does not cleanly shutdown. We get OSError.
            while True:
                response, ela = client.await(connection, timeout=4.0)
                if response:
                    expected = self.enip_server_tcp.config.product_name
                    self.assertEqual(expected, response['enip']['CIP']['list_identity']['CPF']['item'][0]['identity_object']['product_name'])
                else:
                    break

    def test_list_interfaces_tcp(self):
        with client.connector(host=self.enip_server_tcp.addr,
                              port=self.enip_server_tcp.port, timeout=4.0,
                              udp=False, broadcast=False) as conn:
            conn.list_interfaces()
            conn.shutdown()
            while True:
                response, ela = client.await(conn, timeout=4.0)
                if response:
                    self.assertDictEqual({'count': 0}, response['enip']['CIP']['list_interfaces']['CPF'])
                else:
                    break

    def test_list_interfaces_udp(self):
        with client.connector(host=self.enip_server_udp.addr,
                              port=self.enip_server_udp.port, timeout=4.0,
                              udp=True, broadcast=True) as conn:
            conn.list_interfaces()
            # TODO: udp does not cleanly shutdown. We get OSError.
            while True:
                response, ela = client.await(conn, timeout=4.0)
                if response:
                    self.assertDictEqual({'count': 0}, response['enip']['CIP']['list_interfaces']['CPF'])
                else:
                    break

    # Tests related to restart of ENIP device.. 
    # def test_send_NOP(self):
    #     # test tcp
    #     pass

    def test_malformend_request_tcp(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.enip_server_tcp.addr, self.enip_server_tcp.port))
        s.send(b'e\x00\x04\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' +
               b'x00\x00\x01\x00\x00\x00')  # test the help command
        pkt_data = s.recv(1024)
        s.close()
        # TODO: verify data packet?

    def test_malformend_request_udp(self):
        pass


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import os
import sys
import fs
import conpot
from fs import open_fs, errors, subfs
from conpot.core.filesystem import AbstractFS, SubAbstractFS

logger = logging.getLogger(__name__)


class VirtualFS(object):
    """
    Conpot's virtual file system. Based on Pyfilesystem2, it would allow us to have
    arbitrary file uploads while sand boxing them for later analysis. This is how it should look like:

                      [_conpot_vfs]
                            |
                            |-- data_fs (persistent)
                            |    |-- ftp/uploads
                            |    `-- misc.
                            |
                            `-- protocol_fs (temporary, refreshed at startup)
                                 |-- common
                                 |-- telnet
                                 |-- http
                                 |-- snmp
                                 `-- ftp etc.
    :param data_fs_path: Path for storing data_fs. A dictionary with attribute name _protocol_vfs stores all the
    fs folders made by all the individual protocols.
    :type data_fs_path: fs.open_fs
    """
    def __init__(self, data_fs_path=None):
        self._conpot_vfs = dict()   # dictionary to keep all the protocol vfs instances, maintain easy access for
        # individual mounted protocols with paths
        if data_fs_path is None:
            try:
                self.data_fs = open_fs(os.path.join('/'.join(conpot.__file__.split('/')[:-1]), 'tests', 'data',
                                                    'data_temp_fs'))
            except fs.errors.FSError:
                logger.exception('Unable to create persistent storage for Conpot. Exiting')
                sys.exit(3)
        else:
            try:
                assert data_fs_path and isinstance(data_fs_path, str)
                self.data_fs = open_fs(data_fs_path)  # Specify the place where you would place the uploads
            except AssertionError:
                logger.exception('Incorrect FS url specified. Please check documentation for more details.')
                sys.exit(3)
            except fs.errors.CreateFailed:
                logger.exception('Unexpected error occurred while creating Conpot FS.')
                sys.exit(3)
        self.protocol_fs = None

    def initialize_vfs(self, fs_path=None, data_fs_path=None, temp_dir=None):
        if data_fs_path is not None:
            logger.info('Opening path {} for persistent storage of files.'.format(data_fs_path))
            self.__init__(data_fs_path=data_fs_path)
        if fs_path is None:
            fs_path = 'tar://' + os.path.join('/'.join(conpot.__file__.split('/')[:-1]), 'data.tar')
            logger.warning('Using default FS path. {}'.format(fs_path))
        self.protocol_fs = AbstractFS(src_path=fs_path, temp_dir=temp_dir)

    def add_protocol(self,
                     protocol_name: str,
                     data_fs_subdir: str,
                     vfs_dst_path: str,
                     src_path=None,
                     owner_uid=0,
                     group_gid=0,
                     perms=0o755) -> (SubAbstractFS, subfs.SubFS):
        """
        Method that would be used by protocols to initialize vfs. May be called by each protocol individually. This
        creates a chroot jail sub file system env which makes easier handling. It also creates a data_fs sub file system
        for managing protocol specific uploads.
        :param protocol_name: name of the protocol for which VFS is being created.
        :param data_fs_subdir: sub-folder name within data_fs that would be storing the uploads for later analysis
        :param vfs_dst_path:  protocol specific sub-folder path in the fs.
        :param src_path: Source from where the files are to copied.
        :param owner_uid: UID of a registered user. This is the default owner in the sub file system
        :param group_gid: GID of a existing group.
        :param perms: Default permissions of the sub file system.
        :return: fs object

        **Note:** The owner_uid and group_gid must be already registered with the fs. Otherwise an exception
        would be raised.
        """
        assert isinstance(protocol_name, str) and protocol_name
        assert isinstance(data_fs_subdir, str) and data_fs_subdir
        assert isinstance(vfs_dst_path, str) and vfs_dst_path
        if src_path:
            assert isinstance(src_path, str)
            if not os.path.isdir(src_path):
                logger.exception('Protocol directory is not a valid directory.')
                sys.exit(3)
        logger.info('Creating persistent data store for protocol: {}'.format(protocol_name))
        # create a sub directory for persistent storage.
        if self.data_fs.isdir(data_fs_subdir):
            sub_data_fs = self.data_fs.opendir(path=data_fs_subdir)
        else:
            sub_data_fs = self.data_fs.makedir(path=data_fs_subdir)
        if protocol_name not in self._conpot_vfs.keys():
            sub_protocol_fs = self.protocol_fs.mount_fs(vfs_dst_path,
                                                        src_path,
                                                        owner_uid,
                                                        group_gid,
                                                        perms)
            self._conpot_vfs[protocol_name] = (sub_protocol_fs, sub_data_fs)
        return self._conpot_vfs[protocol_name]

    def close(self, force=False):
        """
        Close the filesystem properly. Better and more graceful than __del__
        :param force: Force close. This would close the AbstractFS instance - without close closing data_fs File Systems
        """
        if self._conpot_vfs and (not force):
            for _fs in self._conpot_vfs.keys():
                try:
                    # First let us close all the data_fs instances.
                    self._conpot_vfs[_fs][1].close()
                    # Let us close the protocol_fs sub dirs for that protocol
                    self._conpot_vfs[_fs][0].close()
                except fs.errors.FSError:
                    logger.exception('Error occurred while closing FS {}'.format(_fs))
                    del self._conpot_vfs[_fs][0]
        self.protocol_fs.close()
        self.protocol_fs.clean()


# Copyright (C) 2018 Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from conpot.core import get_interface
from datetime import datetime

core_interface = get_interface()


def conpot_protocol(cls):
    class Wrapper(object):
        def __init__(self, *args, **kwargs):
            self.wrapped = cls(*args, **kwargs)
            self.cls = cls
            if self.cls.__name__ not in 'Proxy':
                core_interface.protocols[self.cls] = self.wrapped

        def __getattr__(self, name):
            if name == 'handle':
                # assuming that handle function from a class is only called when a client tries to connect with an
                # enabled protocol, update the last_active (last_attacked attribute)
                # FIXME: No handle function in HTTPServer
                core_interface.last_active = datetime.now().strftime("%b %d %Y - %H:%M:%S")
            return getattr(core_interface.protocols[self.cls], name)

        def __repr__(self):
            return self.cls.__repr__(self.wrapped)

        __doc__ = property(lambda self: self.cls.__doc__)
        __module__ = property(lambda self: self.cls.__module__)
        __name__ = property(lambda self: self.cls.__name__)
    return Wrapper



# Copyright (C) 2013  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import uuid
from datetime import datetime
from io import StringIO

import unittest
from configparser import ConfigParser
from lxml import etree

from conpot.core.loggers.taxii_log import TaxiiLogger
from conpot.core.loggers.stix_transform import StixTransformer
import sdv.validators as validators
#from conpot.tests.helpers.mitre_stix_validator import STIXValidator


class TestLoggers(unittest.TestCase):

    @unittest.skip('disabled until STIX 2.0')
    def test_stix_transform(self):
        """
        Objective: Test if our STIX xml can be validated.
        """
        config = ConfigParser()
        config_file = os.path.join(os.path.dirname(__file__), '../conpot.cfg')
        config.read(config_file)
        config.set('taxii', 'enabled', True)

        test_event = {'remote': ('127.0.0.1', 54872), 'data_type': 's7comm',
                      'public_ip': '111.222.111.222',
                      'timestamp': datetime.now(),
                      'session_id': str(uuid.uuid4()),
                      'data': {0: {'request': 'who are you', 'response': 'mr. blue'},
                               1: {'request': 'give me apples', 'response': 'no way'}}}
        dom = etree.parse('conpot/templates/default/template.xml')
        stixTransformer = StixTransformer(config, dom)
        stix_package_xml = stixTransformer.transform(test_event)

        validator = validators.STIXSchemaValidator()
        result = validator.validate(StringIO(stix_package_xml.encode('utf-8'))).as_dict()

        has_errors = False
        error_string = ''
        if 'errors' in result:
            has_errors = True
            for error in result['errors']:
                error_string += error['message']
                error_string += ', '
        self.assertFalse(has_errors, 'Error while validations STIX xml: {0}'. format(error_string))

    @unittest.skip('disabled until taxii server is up and running again')
    def test_taxii(self):
        """
        Objective: Test if we can transmit data to MITRE's TAXII test server.
        Note: This actually also tests the StixTransformer since the event is parsed by the transformer
        before transmission.
        """
        config = ConfigParser()
        config_file = os.path.join(os.path.dirname(__file__), '../conpot.cfg')
        config.read(config_file)
        config.set('taxii', 'enabled', True)

        test_event = {'remote': ('127.0.0.1', 54872), 'data_type': 's7comm',
                      'timestamp': datetime.now(),
                      'public_ip': '111.222.111.222',
                      'session_id': str(uuid.uuid4()),
                      'data': {0: {'request': 'who are you', 'response': 'mr. blue'},
                               1: {'request': 'give me apples', 'response': 'no way'}}}
        dom = etree.parse('conpot/templates/default/template.xml')
        taxiiLogger = TaxiiLogger(config, dom)
        taxii_result = taxiiLogger.log(test_event)
        # TaxiiLogger returns false if the message could not be delivered
        self.assertTrue(taxii_result)



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from os import R_OK, W_OK
from lxml import etree
import conpot.core as conpot_core
from gevent.server import StreamServer
from conpot.protocols.ftp.ftp_utils import ftp_commands, FTPException
from conpot.protocols.ftp.ftp_handler import FTPCommandChannel
from conpot.core.protocol_wrapper import conpot_protocol
from datetime import datetime
import logging
logger = logging.getLogger(__name__)
# import sys
# import logging as logger
# logger.basicConfig(stream=sys.stdout, level=logger.DEBUG)


class FTPConfig(object):
    def __init__(self, template):
        self.user_db = dict()  # user_db[uid] = (user_pass, user_group)
        self.grp_db = dict()   # grp_db[gid] = {group: 'group_name'. users: set(users_uid))
        dom = etree.parse(template)
        # First let us get FTP related data
        self.all_commands = ['USER', 'PASS', 'HELP', 'NOOP', 'QUIT', 'SITE HELP', 'SITE', 'SYST', 'TYPE', 'PASV',
                             'PORT', 'ALLO', 'MODE', 'SIZE', 'PWD', 'MKD', 'RMD', 'CWD', 'CDUP', 'MDTM', 'DELE',
                             'SITE CHMOD', 'RNFR', 'RNTO', 'STAT', 'LIST', 'NLST', 'RETR', 'REIN', 'ABOR', 'STOR',
                             'APPE', 'REST', 'STRU', 'STOU']
        # Implementation Note: removing a command from here would make it unrecognizable in FTP server.
        self.enabled_commands = \
            (''.join(dom.xpath('//ftp/device_info/enabled_commands/text()')[0].strip().split())).split(',')
        self.enabled_commands = [i.replace("'", "") for i in self.enabled_commands]
        if 'SITEHELP' in self.enabled_commands:
            self.enabled_commands.remove('SITEHELP')
            self.enabled_commands.append('SITE HELP')
        if 'SITECHMOD' in self.enabled_commands:
            self.enabled_commands.remove('SITECHMOD')
            self.enabled_commands.append('SITE CHMOD')
        for i in self.enabled_commands:
            assert i in self.all_commands
        self.device_type = dom.xpath('//ftp/device_info/device_type/text()')[0]
        self.banner = dom.xpath('//ftp/device_info/banner/text()')[0]
        self.max_login_attempts = int(dom.xpath('//ftp/device_info/max_login_attempts/text()')[0])
        # set the connection timeout to 300 secs.
        self.timeout = int(dom.xpath('//ftp/device_info/sever_timeout/text()')[0])
        if dom.xpath('//ftp/device_info/motd/text()'):
            self.motd = dom.xpath('//ftp/device_info/motd/text()')[0]
        else:
            self.motd = None
        self.stou_prefix = dom.xpath('//ftp/device_info/stou_prefix/text()')
        self.stou_suffix = dom.xpath('//ftp/device_info/stou_suffix/text()')
        # Restrict FTP to only enabled FTP commands
        self.COMMANDS = {i: ftp_commands[i] for i in self.enabled_commands}

        # -- Now that we fetched FTP meta, let us populate users.
        grp = dom.xpath('//ftp/ftp_users/users')[0].attrib['group']
        for i in dom.xpath('//ftp/ftp_users/users/*'):
            self.user_db[int(i.attrib['uid'])] = {
                'uname': i.xpath('./uname/text()')[0],
                'grp': grp,
                'password': i.xpath('./password/text()')[0]
            }
        self.anon_auth = bool(dom.xpath('//ftp/ftp_users/anon_login')[0].attrib['enabled'])
        if self.anon_auth:
            self.anon_uid = int(dom.xpath('//ftp/ftp_users/anon_login')[0].attrib['uid'])
            self.user_db[self.anon_uid] = {
                'uname': dom.xpath('//ftp/ftp_users/anon_login/uname/text()')[0],
                'grp': grp,
                'password': ''
            }

        # As a last step, get VFS related data.
        self.root_path = dom.xpath('//ftp/ftp_vfs/path/text()')[0]
        self.data_fs_subdir = dom.xpath('//ftp/ftp_vfs/data_fs_subdir/text()')[0]
        if len(dom.xpath('//ftp/ftp_vfs/add_src/text()')) == 0:
            self.add_src = None
        else:
            self.add_src = dom.xpath('//ftp/ftp_vfs/add_src/text()')[0].lower()
        # default ftp owners and groups
        self.default_owner = int(dom.xpath('//ftp/ftp_vfs/default_owner/text()')[0])
        self.default_group = int(grp.split(':')[0])

        self.default_perms = oct(int(dom.xpath('//ftp/ftp_vfs/default_perms/text()')[0], 8))
        self.file_default_perms = oct(int(dom.xpath('//ftp/ftp_vfs/upload_file_perms/text()')[0], 8))
        self.dir_default_perms = oct(int(dom.xpath('//ftp/ftp_vfs/upload_file_perms/text()')[0], 8))
        self._custom_files = dom.xpath('//ftp/ftp_vfs/file')
        self._custom_dirs = dom.xpath('//ftp/ftp_vfs/dir')
        self._init_user_db()   # Initialize User DB
        self._init_fs()        # Initialize FTP file system.

    def _init_fs(self):
        # Create/register all necessary users and groups in the file system
        _ = {conpot_core.get_vfs().register_user(uid=k, name=v['uname']) for k, v in self.user_db.items()}
        _ = {conpot_core.get_vfs().create_group(gid=k, name=v['group']) for k, v in self.grp_db.items()}
        _ = {conpot_core.get_vfs().add_users_to_group(gid=k, uids=list(v['users'])) for k, v in self.grp_db.items()}
        # Initialize file system
        self.vfs, self.data_fs = conpot_core.add_protocol(protocol_name='ftp',
                                                          data_fs_subdir=self.data_fs_subdir,
                                                          vfs_dst_path=self.root_path,
                                                          src_path=self.add_src,
                                                          owner_uid=self.default_owner,
                                                          group_gid=self.default_group,
                                                          perms=self.default_perms)
        if self.add_src:
            logger.info('FTP Serving File System from {} at {} in vfs. FTP data_fs sub directory: {}'.format(
                self.add_src, self.root_path, self.data_fs._sub_dir
            ))
        else:
            logger.info('FTP Serving File System at {} in vfs. FTP data_fs sub directory: {}'.format(
                self.root_path, self.data_fs._sub_dir
            ))
        logger.debug('FTP serving list of files : {}'.format(', '.join(self.vfs.listdir('.'))))
        self.root = '/'  # Setup root dir.
        # check for permissions etc.
        logger.debug("FTP root {} is a directory".format(self.vfs.getcwd() + self.root))
        if self.vfs.access(self.root, 0, R_OK):
            logger.debug("FTP root {} is readable".format(self.vfs.getcwd() + self.root))
        else:
            raise FTPException("FTP root must be readable")
        if self.vfs.access(self.root, 0, W_OK):
            logger.debug("FTP root {} is writable".format(self.vfs.getcwd() + self.root))
        else:
            logger.warning("FTP root {} is not writable".format(self.vfs.getcwd() + self.root))
        # Finally apply permissions to specific files.
        for _file in self._custom_files:
            _path = _file.attrib['path']
            _path = _path.replace(self.root_path, self.root)
            _owner = int(_file.xpath('./owner_uid/text()')[0])
            _perms = oct(int(_file.xpath('./perms/text()')[0], 8))
            _accessed = datetime.fromtimestamp(float(_file.xpath('./last_accessed/text()')[0]))
            _modified = datetime.fromtimestamp(float(_file.xpath('./last_modified/text()')[0]))
            self.vfs.chown(_path, _owner, self.default_group)
            self.vfs.chmod(_path, _perms)
            _fs = self.vfs.delegate_fs().delegate_fs()
            _fs.settimes(self.vfs.delegate_path(_path)[1], _accessed, _modified)

        for _dir in self._custom_dirs:
            _path = _dir.attrib['path']
            _recursive = bool(_dir.attrib['recursive'])
            _path = _path.replace(self.root_path, self.root)
            _owner = int(_dir.xpath('./owner_uid/text()')[0])
            _perms = oct(int(_dir.xpath('./perms/text()')[0], 8))
            _accessed = datetime.fromtimestamp(float(_dir.xpath('./last_accessed/text()')[0]))
            _modified = datetime.fromtimestamp(float(_dir.xpath('./last_modified/text()')[0]))
            self.vfs.chown(_path, _owner, self.default_group, _recursive)
            self.vfs.chmod(_path, _perms)
            _fs = self.vfs.delegate_fs().delegate_fs()
            _fs.settimes(self.vfs.delegate_path(_path)[1], _accessed, _modified)
        # self.default_owner = 13
        # self.default_group = 45
        # self.vfs.chmod('/', self.default_perms, recursive=True)
        # self.vfs.chown('/', uid=self.default_owner, gid=self.default_group, recursive=True)

    # --------------------------------------------
    # TODO: move this method to auth module.
    def _init_user_db(self):
        """
        We expect the following dict format to build for every user
                self.user_db[10] = {
                    'uname': 'test_user',
                    'grp': '45:ftp',
                    'password': 'test'
                }
        :return:
        """
        # TODO: Get users from the template.
        self.user_db[13] = {
            'uname': 'nobody',
            'grp': '45:ftp',
            'password': 'nobody'
        }
        # Toggle enable/disable anonymous user.
        self.user_db[22] = {
            'uname': 'anonymous',
            'grp': '45:ftp',
            'password': ''
        }
        # Let us create groups from the populated users.
        for i in self.user_db.keys():
            grp = self.user_db[i].pop('grp')
            _gid, _gname = grp.split(':')
            _gid = int(_gid)
            if _gid not in self.grp_db.keys():
                # It is a new group. Let us create/register this.
                self.grp_db[_gid] = {'group': _gname, 'users': set()}
            self.grp_db[_gid]['users'].add(i)
        # create a simple set of user and pass combinations for easy auth
        self.user_pass = set(zip([v['uname'] for v in self.user_db.values()],
                                 [v['password'] for v in self.user_db.values()]))

    # TODO: move this method to auth module.
    def get_uid(self, user_name):
        """Get uid from a username"""
        [_uid] = [k for k, v in self.user_db.items() if user_name in v.values()]
        return _uid

    # TODO: move this method to auth module.
    def get_gid(self, uid):
        """Get group id of a user from it's uid"""
        [_gid] = [k for k, v in self.grp_db.items() if uid in v['users']]
        return _gid


@conpot_protocol
class FTPServer(object):
    def __init__(self, template, template_directory, args):
        self.template = template
        self.server = None  # Initialize later
        # Initialize vfs here..
        self.handler = FTPCommandChannel
        self.handler.config = FTPConfig(self.template)

    def start(self, host, port):
        self.handler.host, self.handler.port = host, port
        connection = (self.handler.host, self.handler.port)
        self.server = StreamServer(connection, self.handler.stream_server_handle)
        logger.info('FTP server started on: {}'.format(connection))
        self.server.serve_forever()

    def stop(self):
        logger.debug('Stopping FTP server')
        self.server.stop()
        del self.handler


# ---- For debugging ----
if __name__ == '__main__':
    # Set vars for connection information
    TCP_IP = '127.0.0.1'
    TCP_PORT = 10001
    import os
    conpot_core.initialize_vfs()
    test_template = os.getcwd() + '/../../templates/default/ftp/ftp.xml'
    server = FTPServer(test_template, None, None)
    try:
        server.start(TCP_IP, TCP_PORT)
    except KeyboardInterrupt:
        server.stop()


#!/usr/bin/env python
# Copyright (C) 2014 Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import logging
# import uuid
import time

from datetime import datetime

import configparser
from gevent.queue import Empty

from conpot.core.loggers.sqlite_log import SQLiteLogger
# from conpot.core.loggers.mysql_log import MySQLlogger
from conpot.core.loggers.hpfriends import HPFriendsLogger
from conpot.core.loggers.syslog import SysLogger
from conpot.core.loggers.taxii_log import TaxiiLogger
from conpot.core.loggers.json_log import JsonLogger
from .helpers import json_default

logger = logging.getLogger(__name__)


class LogWorker(object):
    def __init__(self, config, dom, session_manager, public_ip):
        self.config = config
        self.log_queue = session_manager.log_queue
        self.session_manager = session_manager
        self.sqlite_logger = None
        # self.mysql_logger = None
        self.json_logger = None
        self.friends_feeder = None
        self.syslog_client = None
        self.public_ip = public_ip
        self.taxii_logger = None

        if config.getboolean('sqlite', 'enabled'):
            self.sqlite_logger = SQLiteLogger()

        # if config.getboolean('mysql', 'enabled'):
        #     host = config.get('mysql', 'host')
        #     port = config.getint('mysql', 'port')
        #     db = config.get('mysql', 'db')
        #     username = config.get('mysql', 'username')
        #     passphrase = config.get('mysql', 'passphrase')
        #     logdevice = config.get('mysql', 'device')
        #     logsocket = config.get('mysql', 'socket')
        #     sensorid = config.get('common', 'sensorid')
        #     self.mysql_logger = MySQLlogger(host, port, db, username, passphrase, logdevice, logsocket, sensorid)

        if config.getboolean('json', 'enabled'):
            filename = config.get('json', 'filename')
            sensorid = config.get('common', 'sensorid')
            self.json_logger = JsonLogger(filename, sensorid, public_ip)

        if config.getboolean('hpfriends', 'enabled'):
            host = config.get('hpfriends', 'host')
            port = config.getint('hpfriends', 'port')
            ident = config.get('hpfriends', 'ident')
            secret = config.get('hpfriends', 'secret')
            channels = eval(config.get('hpfriends', 'channels'))
            try:
                self.friends_feeder = HPFriendsLogger(host, port, ident, secret, channels)
            except Exception as e:
                logger.exception(e.message)
                self.friends_feeder = None

        if config.getboolean('syslog', 'enabled'):
            host = config.get('syslog', 'host')
            port = config.getint('syslog', 'port')
            facility = config.get('syslog', 'facility')
            logdevice = config.get('syslog', 'device')
            logsocket = config.get('syslog', 'socket')
            self.syslog_client = SysLogger(host, port, facility, logdevice, logsocket)

        if config.getboolean('taxii', 'enabled'):
            # TODO: support for certificates
            self.taxii_logger = TaxiiLogger(config, dom)

        self.enabled = True

    def _process_sessions(self):
        sessions = self.session_manager._sessions
        try:
            session_timeout = self.config.get("session", "timeout")
        except (configparser.NoSectionError, configparser.NoOptionError):
            session_timeout = 5
        for session in sessions:
            if len(session.data) > 0:
                sec_last_event = max(session.data) / 1000
            else:
                sec_last_event = 0
            sec_session_start = time.mktime(session.timestamp.timetuple())
            sec_now = time.mktime(datetime.utcnow().timetuple())
            if (sec_now - (sec_session_start + sec_last_event)) >= float(session_timeout):
                # TODO: We need to close sockets in this case
                logger.info('Session timed out: %s', session.id)
                session.set_ended()
                sessions.remove(session)

    def start(self):
        self.enabled = True
        while self.enabled:
            try:
                event = self.log_queue.get(timeout=2)
            except Empty:
                self._process_sessions()
            else:
                if self.public_ip:
                    event["public_ip"] = self.public_ip

                if self.friends_feeder:
                    self.friends_feeder.log(json.dumps(event, default=json_default))

                if self.sqlite_logger:
                    self.sqlite_logger.log(event)

                # if self.mysql_logger:
                #     self.mysql_logger.log(event)

                if self.syslog_client:
                    self.syslog_client.log(event)

                if self.taxii_logger:
                    self.taxii_logger.log(event)

                if self.json_logger:
                    self.json_logger.log(event)

    def stop(self):
        self.enabled = False



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
from . import messages
import copy
from lxml import etree

from .register import KamstrupRegister


logger = logging.getLogger(__name__)


class CommandResponder(object):
    def __init__(self, template):
        # key: kamstrup_meter register, value: databus key
        self.registers = {}

        dom = etree.parse(template)
        registers = dom.xpath('//kamstrup_meter/registers/*')
        self.communication_address = int(dom.xpath('//kamstrup_meter/config/communication_address/text()')[0])
        for register in registers:
            name = int(register.attrib['name'])
            length = int(register.attrib['length'])
            units = int(register.attrib['units'])
            unknown = int(register.attrib['unknown'])
            databuskey = register.xpath('./value/text()')[0]
            kamstrup_register = KamstrupRegister(name, units, length, unknown, databuskey)
            assert name not in self.registers
            self.registers[name] = kamstrup_register

    def respond(self, request):
        if request.communication_address != self.communication_address:
            logger.warning('Kamstrup request received with wrong communication address, got {} but expected {}.'.format(
                request.communication_address, self.communication_address
            ))
            return None
        elif isinstance(request, messages.KamstrupRequestGetRegisters):
            response = messages.KamstrupResponseRegister(self.communication_address)
            for register in request.registers:
                if register in self.registers:
                    response.add_register(copy.deepcopy(self.registers[register]))
            return response
        else:
            assert False



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


class KamstrupRegister(object):
    def __init__(self, name, units, length, unknown, databus_key):
        self.name = name
        self.units = units
        self.length = length
        self.unknown = unknown
        self.databus_key = databus_key



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
from conpot.helpers import chr_py3
import crc16

from . import kamstrup_constants
from .messages import KamstrupRequestGetRegisters, KamstrupRequestUnknown

logger = logging.getLogger(__name__)


class KamstrupRequestParser(object):
    def __init__(self):
        self.bytes = list()
        self.parsing = False
        self.data_escaped = False
        self.done = False
        self.request_map = {KamstrupRequestGetRegisters.command_byte: KamstrupRequestGetRegisters}

    def add_byte(self, byte):
        self.bytes.append(ord(byte))

    def get_request(self):
        bytes_len = len(self.bytes)
        position = 0
        while position < bytes_len:
            d = self.bytes[position]
            if not self.parsing and d != kamstrup_constants.REQUEST_MAGIC:
                logger.info('Kamstrup skipping byte, expected kamstrup_meter request magic but got: {0}'
                            .format(hex(d)))
                del self.bytes[position]
                bytes_len -= 1
                continue
            else:
                self.parsing = True

                escape_escape_byte = False
                if self.data_escaped:
                    self.bytes[position] ^= 0xff
                    if d is kamstrup_constants.EOT_MAGIC:
                        escape_escape_byte = True
                    self.data_escaped = False
                elif d is kamstrup_constants.ESCAPE:
                    self.data_escaped = True
                    del self.bytes[position]
                    bytes_len -= 1
                    continue

                assert self.data_escaped is False

                if d is kamstrup_constants.EOT_MAGIC and not escape_escape_byte:
                    if not self.valid_crc(self.bytes[1:position]):
                        self.parsing = False
                        del self.bytes[0:position]
                        logger.warning('Kamstrup CRC check failed for request.')
                    # now we expect (0x80, 0x3f, 0x10) =>
                    # (request magic, communication address, command byte)
                    comm_address = self.bytes[1]
                    command_byte = self.bytes[2]
                    if self.bytes[2] in self.request_map:
                        result = self.request_map[command_byte](comm_address, command_byte, self.bytes[3:-3])
                        del self.bytes[:position + 1]
                    else:
                        result = KamstrupRequestUnknown(comm_address, command_byte, self.bytes[3:-3])
                        del self.bytes[:position + 1]
                    return result
                position += 1

    @classmethod
    def valid_crc(cls, message):
        supplied_crc = message[-2] * 256 + message[-1]
        calculated_crc = crc16.crc16xmodem(b''.join([chr_py3(item) for item in message[:-2]]))
        return supplied_crc == calculated_crc



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging

import gevent
from gevent import event
import conpot.core as conpot_core


logger = logging.getLogger(__name__)


# Simulates power usage for a Kamstrup 382 meter
class UsageSimulator(object):
    def __init__(self, *args):
        self._enabled = True
        self.stopped = gevent.event.Event()
        # both highres, lowres will be calculated on the fly
        self.energy_in = 0
        self.energy_out = 0
        # p1, p2, p3
        self.voltage = [0, 0, 0]
        self.current = [0, 0, 0]
        self.power = [0, 0, 0]
        gevent.spawn(self.initialize)

    def usage_counter(self):
        while self._enabled:
            # since this is gevent, this actually sleep for _at least_ 1 second
            # TODO: measure last entry and figure it out < jkv: Figure what out?!?
            gevent.sleep(1)
            for x in [0, 1, 2]:
                self.energy_in += int(self.power[x] * 0.0036)
        # ready for shutdown!
        self.stopped.set()

    def stop(self):
        self._enabled = False
        self.stopped.wait()

    def initialize(self):
        # we need the databus initialized before we can probe values
        databus = conpot_core.get_databus()
        databus.initialized.wait()

        # accumulated counter
        energy_in_register = 'register_13'
        self.energy_in = databus.get_value(energy_in_register)
        databus.set_value(energy_in_register, self._get_energy_in)
        databus.set_value('register_1', self._get_energy_in_lowres)

        energy_out_register = 'register_14'
        self.energy_out = databus.get_value(energy_out_register)
        databus.set_value(energy_out_register, self._get_energy_out)
        databus.set_value('register_2', self._get_energy_out_lowres)

        volt_1_register = 'register_1054'
        self.voltage[0] = databus.get_value(volt_1_register)
        databus.set_value(volt_1_register, self._get_voltage_1)

        volt_2_register = 'register_1055'
        self.voltage[1] = databus.get_value(volt_2_register)
        databus.set_value(volt_2_register, self._get_voltage_2)

        volt_3_register = 'register_1056'
        self.voltage[2] = databus.get_value(volt_3_register)
        databus.set_value(volt_3_register, self._get_voltage_3)

        current_1_register = 'register_1076'
        self.current[0] = databus.get_value(current_1_register)
        databus.set_value(current_1_register, self._get_current_1)

        current_2_register = 'register_1077'
        self.current[1] = databus.get_value(current_2_register)
        databus.set_value(current_2_register, self._get_current_2)

        current_3_register = 'register_1078'
        self.current[2] = databus.get_value(current_3_register)
        databus.set_value(current_3_register, self._get_current_3)

        power_1_register = 'register_1080'
        self.power[0] = databus.get_value(power_1_register)
        databus.set_value(power_1_register, self._get_power_1)

        power_2_register = 'register_1081'
        self.power[1] = databus.get_value(power_2_register)
        databus.set_value(power_2_register, self._get_power_2)

        power_3_register = 'register_1082'
        self.power[2] = databus.get_value(power_3_register)
        databus.set_value(power_3_register, self._get_power_3)

        gevent.spawn(self.usage_counter)

    def _get_energy_in(self):
        return self.energy_in

    def _get_energy_out(self):
        return self.energy_out

    def _get_energy_in_lowres(self):
        return self.energy_in / 1000

    def _get_energy_out_lowres(self):
        return self.energy_out / 1000

    def _get_voltage_1(self):
        return self.voltage[0]

    def _get_voltage_2(self):
        return self.voltage[1]

    def _get_voltage_3(self):
        return self.voltage[2]

    def _get_current_1(self):
        return self.current[0]

    def _get_current_2(self):
        return self.current[1]

    def _get_current_3(self):
        return self.current[2]

    def _get_power_1(self):
        return self.power[0]

    def _get_power_2(self):
        return self.power[1]

    def _get_power_3(self):
        return self.power[2]



# This implementation of the S7 protocol is highly inspired
# by the amazing plcscan work by the ScadaStrangeLove group.
# https://code.google.com/p/plcscan/source/browse/trunk/s7.py

from struct import *
import struct
from conpot.helpers import str_to_bytes
from conpot.protocols.s7comm.exceptions import ParseException


class COTP(object):
    def __init__(self, tpdu_type=0, opt_field=0, payload='', trailer=''):
        self.tpdu_type = tpdu_type
        self.opt_field = opt_field
        self.payload = payload
        self.trailer = trailer

        if self.tpdu_type == 240:
            self.packet_length = 2
        else:
            self.packet_length = 1 + len(self.payload)

            # COTP BASE PACKET FORMAT:
            # -------------------------------------
            #           1 byte      LENGTH (=n + 1)
            #           1 byte      TPDU TYPE
            #           1 byte      OPT FIELD (optional!), bitmask!
            #           n bytes     TPDU PAYLOAD
            #           x bytes     TRAILER (optional!), most probably containing S7.

    def pack(self):
        if self.tpdu_type == 0xf0:
            return pack('!BBB', self.packet_length, self.tpdu_type, self.opt_field) + str_to_bytes(self.payload) + \
                   str_to_bytes(self.trailer)
        else:
            return pack('!BB', self.packet_length, self.tpdu_type) + str_to_bytes(self.payload) +\
                   str_to_bytes(self.trailer)

    def parse(self, packet):

        try:
            header = unpack('!BBB', packet[:3])
        except struct.error:
            raise ParseException('s7comm', 'malformed packet header structure')

        self.packet_length = header[0]
        self.tpdu_type = int(header[1])
        self.trailer = packet[1 + self.packet_length:]

        if self.tpdu_type == 0xf0:
            # the DT DATA TPDU features another header byte that shifts our structure
            self.opt_field = header[2]
            self.payload = packet[3:1 + self.packet_length]
        else:
            self.payload = packet[2:1 + self.packet_length]

        return self


# COTP Connection Request or Connection Confirm packet (ISO on TCP). RFC 1006
class COTPConnectionPacket:
    def __init__(self, dst_ref=0, src_ref=0, opt_field=0, src_tsap=0, dst_tsap=0, tpdu_size=0):
        self.dst_ref = dst_ref
        self.src_ref = src_ref
        self.opt_field = opt_field
        self.src_tsap = src_tsap
        self.dst_tsap = dst_tsap
        self.tpdu_size = tpdu_size

        # COTP CR PACKET FORMAT:
        # -------------------------------------
        #           2 bytes     DST REFERENCE
        #           2 bytes     SRC REFERENCE
        #           1 byte      OPTION FIELD (bitmask!)
        #          ---------------------------------------
        #           n bytes     1 byte  PARAM CODE
        #                       1 byte  PARAM LENGTH (n)
        #                       n bytes PARAM DATA
        #          ---------------------------------------
        #           "n" Block repeats until end of packet

    def dissect(self, packet):

        # dissect fixed header
        try:
            fixed_header = unpack('!HHB', packet[:5])
        except struct.error:
            raise ParseException('s7comm', 'malformed fixed header structure')

        self.dst_ref = fixed_header[0]
        self.src_ref = fixed_header[1]
        self.opt_field = fixed_header[2]

        # dissect variable header
        chunk = packet[5:]
        while len(chunk) > 0:
            chunk_param_header = unpack('!BB', chunk[:2])
            chunk_param_code = int(chunk_param_header[0])
            chunk_param_length = chunk_param_header[1]

            if chunk_param_length == 1:
                param_unpack_structure = '!B'
            elif chunk_param_length == 2:
                param_unpack_structure = '!H'
            else:
                raise ParseException('s7comm', 'malformed variable header structure')

            chunk_param_data = unpack(param_unpack_structure, chunk[2:2 + chunk_param_length])

            if chunk_param_code == 0xc1:
                self.src_tsap = chunk_param_data[0]
            elif chunk_param_code == 0xc2:
                self.dst_tsap = chunk_param_data[0]
            elif chunk_param_code == 0xc0:
                self.tpdu_size = chunk_param_data[0]
            else:
                raise ParseException('s7comm', 'unknown parameter code')

            # remove this part of the chunk
            chunk = chunk[2 + chunk_param_length:]

        return self


class COTP_ConnectionConfirm(COTPConnectionPacket):
    def __init__(self, dst_ref=0, src_ref=0, opt_field=0, src_tsap=0, dst_tsap=0, tpdu_size=0):
        self.dst_ref = dst_ref
        self.src_ref = src_ref
        self.opt_field = opt_field
        self.src_tsap = src_tsap
        self.dst_tsap = dst_tsap
        self.tpdu_size = tpdu_size
        super(COTPConnectionPacket, self).__init__()

    def assemble(self):
        return pack('!HHBBBHBBH', self.dst_ref, self.src_ref, self.opt_field,
                    0xc1,  # param code:   src-tsap
                    0x02,  # param length: 2 bytes
                    self.src_tsap,
                    0xc2,  # param code:   dst-tsap
                    0x02,  # param length: 2 bytes
                    self.dst_tsap)


class COTP_ConnectionRequest(COTPConnectionPacket):
    def __init__(self, dst_ref=0, src_ref=0, opt_field=0, src_tsap=0, dst_tsap=0, tpdu_size=0):
        self.dst_ref = dst_ref
        self.src_ref = src_ref
        self.opt_field = opt_field
        self.src_tsap = src_tsap
        self.dst_tsap = dst_tsap
        self.tpdu_size = tpdu_size
        super(COTPConnectionPacket, self).__init__()

    def assemble(self):
        return pack('!HHBBBHBBHBBB', self.dst_ref, self.src_ref, self.opt_field,
                    0xc1,  # param code:   src-tsap
                    0x02,  # param length: 2 bytes
                    self.src_tsap,
                    0xc2,  # param code:   dst-tsap
                    0x02,  # param length: 2 bytes
                    self.dst_tsap,
                    0xc0,  # param code:   tpdu-size
                    0x01,  # param length: 1 byte
                    self.tpdu_size)



from gevent import monkey; monkey.patch_all()
import fs
import os
import logging
import tftpy
import time
from gevent import socket
from fs import errors
from tftpy import TftpException, TftpErrors
from tftpy.TftpStates import TftpStateExpectACK, TftpStateExpectDAT
from tftpy.TftpPacketTypes import TftpPacketRRQ, TftpPacketWRQ
from conpot.helpers import sanitize_file_name

logger = logging.getLogger(__name__)


class TFTPState(tftpy.TftpStates.TftpState):
    def __int__(self, context):
        super(TFTPState, self).__init__(context)

    def handle(self, pkt, raddress, rport):
        raise NotImplementedError


class TFTPServerState(TFTPState):
    """The base class for server states. """
    # We had to rewrite the because -- had to check os.* wrappers.
    vfs, data_fs = None, None
    full_path = None

    def handle(self, pkt, raddress, rport):
        raise NotImplementedError

    def serverInitial(self, pkt, raddress, rport):
        options = pkt.options
        sendoack = False
        if not self.context.tidport:
            self.context.tidport = rport
            logger.info("Setting tidport to %s" % rport)

        logger.debug("Setting default options, blksize")
        self.context.options = {'blksize': tftpy.DEF_BLKSIZE}

        if options:
            logger.debug("Options requested: %s", options)
            supported_options = self.returnSupportedOptions(options)
            self.context.options.update(supported_options)
            sendoack = True
        # FIXME - only octet mode is supported at this time.
        if pkt.mode != 'octet':
            logger.info("Received non-octet mode request. Replying with binary data.")

        # test host/port of client end
        if self.context.host != raddress or self.context.port != rport:
            self.sendError(TftpErrors.UnknownTID)
            logger.error("Expected traffic from %s:%s but received it from %s:%s instead." % (self.context.host,
                                                                                              self.context.port,
                                                                                              raddress,
                                                                                              rport))
            # Return same state, we're still waiting for valid traffic.
            return self
        logger.debug("Requested filename is %s", pkt.filename)
        if pkt.filename.startswith(self.context.root.encode()):
            full_path = pkt.filename
        else:
            full_path = os.path.join(self.context.root, pkt.filename.decode().lstrip('/'))
        try:
            logger.info("Full path of file to be uploaded is {}".format(full_path))
            self.full_path = full_path
        except fs.errors.FSError:
            logger.warning("requested file is not within the server root - bad")
            self.sendError(TftpErrors.IllegalTftpOp)
            raise TftpException("Bad file path")
        self.context.file_to_transfer = pkt.filename
        return sendoack


class TFTPStateServerRecvRRQ(TFTPServerState):
    def handle(self, pkt, raddress, rport):
        """Handle an initial RRQ packet as a server."""
        logger.debug("In TftpStateServerRecvRRQ.handle")
        sendoack = self.serverInitial(pkt, raddress, rport)
        path = self.full_path
        logger.info("Opening file %s for reading" % path)
        if self.context.vfs.norm_path(path):
            self.context.fileobj = self.context.vfs.open(path.replace(self.context.root + '/', ''), "rb")
        else:
            logger.info("File not found: %s", path.replace(self.context.root + '/', ''))
            self.sendError(TftpErrors.FileNotFound)
            raise TftpException("File not found: {}".format(path))

        # Options negotiation.
        if sendoack and 'tsize' in self.context.options:
            # getting the file size for the tsize option. As we handle
            # file-like objects and not only real files, we use this seeking
            # method instead of asking the OS
            self.context.fileobj.seek(0, os.SEEK_END)
            tsize = str(self.context.fileobj.tell())
            self.context.fileobj.seek(0, 0)
            self.context.options['tsize'] = tsize

        if sendoack:
            # Note, next_block is 0 here since that's the proper
            # acknowledgement to an OACK.
            # FIXME: perhaps we do need a TftpStateExpectOACK class...
            self.sendOACK()
            # Note, self.context.next_block is already 0.
        else:
            self.context.next_block = 1
            logger.debug("No requested options, starting send...")
            self.context.pending_complete = self.sendDAT()
        # Note, we expect an ack regardless of whether we sent a DAT or an
        # OACK.
        return TftpStateExpectACK(self.context)

        # Note, we don't have to check any other states in this method, that's
        # up to the caller.


class TFTPStateServerRecvWRQ(TFTPServerState):
    """This class represents the state of the TFTP server when it has just
    received a WRQ packet."""
    def make_subdirs(self):
        """The purpose of this method is to, if necessary, create all of the
        subdirectories leading up to the file to the written."""
        # Pull off everything below the root.
        subpath = self.full_path[len(self.context.root):]
        subpath = subpath.decode() if isinstance(subpath, bytes) else subpath
        logger.debug("make_subdirs: subpath is %s", subpath)
        # Split on directory separators, but drop the last one, as it should
        # be the filename.
        dirs = subpath.split('/')[:-1]
        logger.debug("dirs is %s", dirs)
        current = self.context.root
        for dir in dirs:
            if dir:
                current = os.path.join(current, dir)
                if self.context.vfs.isdir(current):
                    logger.debug("%s is already an existing directory", current)
                else:
                    self.context.vfs.makedir(current, 0o700)

    def handle(self, pkt, raddress, rport):
        """Handle an initial WRQ packet as a server."""
        logger.debug("In TFTPStateServerRecvWRQ.handle")
        sendoack = self.serverInitial(pkt, raddress, rport)
        path = self.full_path
        self.context.file_path = path
        path = path.decode() if isinstance(path, bytes) else path
        logger.info("Opening file %s for writing" % path)
        if self.context.vfs.exists(path):
            logger.warning("File %s exists already, overwriting..." % (
                self.context.file_to_transfer))
        self.make_subdirs()
        self.context.fileobj = self.context.vfs.open(path, "wb")
        # Options negotiation.
        if sendoack:
            logger.debug("Sending OACK to client")
            self.sendOACK()
        else:
            logger.debug("No requested options, expecting transfer to begin...")
            self.sendACK()
        self.context.next_block = 1
        return TftpStateExpectDAT(self.context)


class TFTPStateServerStart(TFTPState):
    """The start state for the server. This is a transitory state since at
    this point we don't know if we're handling an upload or a download. We
    will commit to one of them once we interpret the initial packet."""
    def handle(self, pkt, raddress, rport):
        """Handle a packet we just received."""
        logger.debug("Using TFTPStateServerStart.handle")
        if isinstance(pkt, TftpPacketRRQ):
            logger.debug("Handling an RRQ packet")
            return TFTPStateServerRecvRRQ(self.context).handle(pkt, raddress, rport)
        elif isinstance(pkt, TftpPacketWRQ):
            logger.debug("Handling a WRQ packet")
            return TFTPStateServerRecvWRQ(self.context).handle(pkt, raddress, rport)
        else:
            self.sendError(tftpy.TftpErrors.IllegalTftpOp)
            raise TftpException("Invalid packet to begin upload/download: %s" % pkt)


class TFTPContextServer(tftpy.TftpContexts.TftpContextServer):
    """Simple TFTP server handler wrapper. Use conpot's filesystem wrappers rather than os.*"""
    file_path = None
    _already_uploaded = False  # Since with UDP, we can't differentiate between when a user disconnected
    # after successful upload and when the client timed out, we would allow file copy on data_fs only once

    def __int__(self, host, port, timeout, root, dyn_file_func, upload_open):
        tftpy.TftpContexts.TftpContextServer.__init__(self, host=host, port=port, timeout=timeout, root=root,
                                                      dyn_file_func=None, upload_open=None)
        self.state = TFTPStateServerStart(self)
        self.log = logger
        self.state.log = logger
        self.root = root
        if self.sock:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.bind('', 0)
        self.data_fs_fileobj = None
        self.vfs = None
        self.data_fs = None
        self.sock.setblocking(False)

    def start(self, buffer):
        logger.debug('In TFTPContextServer - Starting TFTP context with : {}'.format(buffer))
        self.metrics.start_time = time.time()
        self.last_update = time.time()
        pkt = self.factory.parse(buffer)
        self.state = TFTPStateServerStart(self)
        self.state = self.state.handle(pkt, self.host, self.port)

    def end(self):
        logger.debug("In TFTPContextServer.end - closing socket and files.")
        self.sock.close()
        if self.fileobj is not None and not self.fileobj.closed:
            logger.debug("self.fileobj is open - closing")
            self.fileobj.close()
        if not self.state and (not self._already_uploaded):
            if self.file_path:
                # Return None only when transfer is complete!
                logger.info('TFTP : Transfer Complete!')
                _file_path = self.file_path if isinstance(self.file_path, str) else self.file_path.decode()
                _data_fs_filename = sanitize_file_name(''.join(_file_path.split('/')[-1:]), self.host, self.port)
                logger.info('Opening {} for data_fs writing.'.format(_data_fs_filename))
                with self.vfs.open(_file_path, "rb") as _vfs_file:
                    with self.data_fs.open(_data_fs_filename, "wb") as _data_file:
                        content = _vfs_file.read()
                        _data_file.write(content)
            self._already_uploaded = True
        self.metrics.end_time = time.time()
        logger.debug("Set metrics.end_time to %s", self.metrics.end_time)
        self.metrics.compute()



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Author: Abhinav Saxena <xandfury@gmail.com>
# Institute of Informatics and Communication, University of Delhi, South Campus.

import gevent
import os
from gevent import monkey; gevent.monkey.patch_all()
from lxml import etree
from conpot.protocols.tftp import tftp_handler
from gevent.server import DatagramServer
import conpot.core as conpot_core
from conpot.core.protocol_wrapper import conpot_protocol
from gevent import event
from tftpy import TftpException, TftpTimeout
import logging
logger = logging.getLogger(__name__)

# For debugging --
# import logging as logger
# import sys
# logger.basicConfig(stream=sys.stdout, level=logger.DEBUG)


@conpot_protocol
class TftpServer(object):
    """TFTP Server"""
    TIMEOUT_RETRIES = 5

    def __init__(self, template, template_directory, args, timeout=5):
        self.timeout = float(timeout)
        self.server = None   # server attr - Initialize in start
        self.root = None
        self.listener = None   # listener socket
        # A dict of sessions, where each session is keyed by a string like
        # ip:tid for the remote end.
        self.sessions = {}
        # A threading event to help threads synchronize with the server is_running state.
        self.is_running = gevent.event.Event()

        self.shutdown = False
        self._init_vfs(template)
        logger.debug('TFTP server initialized.')

    def _init_vfs(self, template):
        dom = etree.parse(template)
        self.root_path = dom.xpath('//tftp/tftp_root_path/text()')[0].lower()
        if len(dom.xpath('//tftp/add_src/text()')) == 0:
            self.add_src = None
        else:
            self.add_src = dom.xpath('//tftp/add_src/text()')[0].lower()
        self.data_fs_subdir = dom.xpath('//tftp/data_fs_subdir/text()')[0].lower()
        # Create a file system.
        self.vfs, self.data_fs = conpot_core.add_protocol(protocol_name='tftp',
                                                          data_fs_subdir=self.data_fs_subdir,
                                                          vfs_dst_path=self.root_path,
                                                          src_path=self.add_src)
        if self.add_src:
            logger.info('TFTP Serving File System from {} at {} in vfs. TFTP data_fs sub directory: {}'.format(
                self.add_src, self.root_path, self.data_fs._sub_dir
            ))
        else:
            logger.info('TFTP Serving File System at {} in vfs. TFTP data_fs sub directory: {}'.format(
                self.root_path, self.data_fs._sub_dir
            ))
        logger.debug('TFTP serving list of files : {}'.format(', '.join(self.vfs.listdir('.'))))
        self.root = '/'  # Setup root dir.
        # check for permissions etc.
        logger.debug("TFTP root {} is a directory".format(self.vfs.getcwd() + self.root))
        if self.vfs.access(self.root, 0, os.R_OK):
            logger.debug("TFTP root {} is readable".format(self.vfs.getcwd() + self.root))
        else:
            raise TftpException("The TFTP root must be readable")
        if self.vfs.access(self.root, 0, os.W_OK):
            logger.debug("TFTP root {} is writable".format(self.vfs.getcwd() + self.root))
        else:
            logger.warning("The TFTP root {} is not writable".format(self.vfs.getcwd() + self.root))

    def handle(self, buffer, client_addr):
        session = conpot_core.get_session('tftp', client_addr[0], client_addr[1])
        logger.info('New TFTP client has connected. Connection from {}:{}. '.format(client_addr[0], client_addr[1]))
        session.add_event({'type': 'NEW_CONNECTION'})
        logger.debug("Read %d bytes", len(buffer))
        context = tftp_handler.TFTPContextServer(client_addr[0], client_addr[1], self.timeout, self.root, None, None)
        context.vfs, context.data_fs = self.vfs, self.data_fs
        if self.shutdown:
            logger.info("Shutting down now. Disconnecting {}".format(client_addr))
            session.add_event({'type': 'CONNECTION_TERMINATED'})
        try:
            context.start(buffer)
            context.cycle()
        except TftpTimeout as err:
            logger.info("Timeout occurred %s: %s" % (context, str(err)))
            session.add_event({'type': 'CONNECTION_TIMEOUT'})
            context.retry_count += 1
            # TODO: We should accept retries from the user.
            if context.retry_count >= self.TIMEOUT_RETRIES:
                logger.info("TFTP: Hit max {} retries on {}, giving up".format(self.TIMEOUT_RETRIES, context))
            else:
                logger.info("TFTP: resending on session %s" % context)
                context.state.resendLast()
        except TftpException as err:
            logger.info("TFTP: Fatal exception thrown from session {}: {}".format(context, str(err)))
            session.add_event({'type': 'CONNECTION_LOST'})
        logger.info('TFTP: terminating connection: {}'.format(context))
        session.set_ended()
        context.end()
        # Gathering up metrics before terminating the connection.
        metrics = context.metrics
        if metrics.duration == 0:
            logger.info("Duration too short, rate undetermined")
        else:
            logger.info("Transferred %d bytes in %.2f seconds" % (metrics.bytes, metrics.duration))
            logger.info("Average rate: %.2f kbps" % metrics.kbps)
        logger.info("%.2f bytes in resent data" % metrics.resent_bytes)
        logger.info("%d duplicate packets" % metrics.dupcount)
        del context

    def start(self, host, port):
        conn = (host, port)
        # FIXME - sockets should be non-blocking
        self.listener = gevent.socket.socket(gevent.socket.AF_INET, gevent.socket.SOCK_DGRAM)
        self.listener.bind(conn)
        self.listener.settimeout(self.timeout)
        self.server = DatagramServer(self.listener, self.handle)
        logger.info('Starting TFTP server at {}'.format(conn))
        self.server.serve_forever()

    def stop(self):
        self.server.close()


if __name__ == '__main__':
    import conpot
    # initialize the file system.
    conpot_core.initialize_vfs()
    # get the current directory
    dir_name = os.path.dirname(conpot.__file__)
    server = TftpServer(dir_name + '/templates/default/tftp/tftp.xml', dir_name + '/templates/default', args=None)
    try:
        server.start('127.0.0.1', 6090)
    except KeyboardInterrupt:
        server.stop()


# Copyright (C) 2013  Daniel creo Haslinger <creo-conpot@blackmesa.at>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from logging.handlers import SysLogHandler
import logging
import socket


class SysLogger(object):
    def __init__(self, host, port, facility, logdevice, logsocket):
        logger = logging.getLogger()

        if str(logsocket).lower() == 'udp':
            logger.addHandler(SysLogHandler(address=(host, port),
                                            facility=getattr(SysLogHandler, 'LOG_' + str(facility).upper()),
                                            socktype=socket.SOCK_DGRAM))
        elif str(logsocket).lower() == 'dev':
            logger.addHandler(SysLogHandler(logdevice))

    def log(self, data):
        # stub function since the additional handler has been added to the root loggers instance.
        pass



# Copyright (C) 2015  Adarsh Dinesh <adarshdinesh@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
import conpot.utils.mac_addr as mac_addr
import subprocess


class TestMacAddrUtil(unittest.TestCase):

    def setUp(self):
        self.change_mac_process = subprocess.Popen(["ip", "li", "delete", "dummy", "type", "dummy"],
                                                   stderr=subprocess.STDOUT, stdout=subprocess.PIPE)

    def tearDown(self):
        self.change_mac_process.terminate()

    @unittest.skip('shunt to a later phase')
    def test_mac(self):
        """
        Objective: Test if the spoofer is able to change MAC address
        """
        testmac = b'00:de:ad:be:ef:00'
        iface = b'dummy'
        # Load dummy module
        s = subprocess.Popen(["modprobe", "dummy"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        # Check if dummy is loaded 
        data = s.stdout.read()        
        if data:
            self.skipTest("Can't create dummy device")            
        # Create a dummy network interface
        subprocess.Popen(["ip", "li", "add", "dummy", "type", "dummy"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        s = subprocess.Popen(["ip", "link", "show"], stdout=subprocess.PIPE)
        data = s.stdout.read()
        if b'dummy' in data:
            # Change mac address of dummy interface and test it
            mac_addr.change_mac(iface=iface, mac=testmac)
            flag = mac_addr._check_mac(iface, testmac)
            # Remove the dummy interface
            with self.change_mac_process:
                self.assertTrue(flag is True)
        else:
            self.skipTest("Can't change MAC address")


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2015  Danilo Massa
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import json
from .helpers import json_default


class JsonLogger(object):

    def __init__(self, filename, sensorid, public_ip):
        self.fileHandle = open(filename, 'a')
        self.sensorid = sensorid
        self.public_ip = public_ip

    def log(self, event):
        data = {
            'timestamp': event['timestamp'].isoformat(),
            'sensorid': self.sensorid,
            'id': event["id"],
            'src_ip': event["remote"][0],
            'src_port': event["remote"][1],
            'dst_ip': self.public_ip,
            'data_type': event["data_type"],
            'request': event["data"].get('request'),
            'response': event["data"].get('response'),
            'event_type': event["data"].get('type'),
        }

        json.dump(data, self.fileHandle, default=json_default)
        self.fileHandle.write("\n")
        self.fileHandle.flush()

    def log_session(self, session):
        pass



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent
from gevent import monkey; gevent.monkey.patch_all()
import unittest
from gevent import socket
import os
import conpot
import conpot.core as conpot_core
from collections import namedtuple
import re
from conpot.protocols.guardian_ast.guardian_ast_server import GuardianASTServer


DATA = {
    'I20100': b'\nI20100\n05/30/2018 19:15\n\nSTATOIL STATION\n\n\n\nIN-TANK INVENTORY\n\nTANK PRODUCT             VOLUME TC VOLUME   ULLAGE   HEIGHT    WATER     TEMP\n  1  SUPER                 2428      2540     4465    39.88     6.62    53.74\n  2  UNLEAD                7457      7543     7874    65.59     8.10    58.17\n  3  DIESEL                6532      6664     4597    33.06     5.91    57.91\n  4  PREMIUM               2839      2867     4597    66.57     4.49    57.88\n',
    'I20200': b'\nI20200\n05/30/2018 19:17\n\nSTATOIL STATION\n\n\n\nDELIVERY REPORT\n\nT 1:SUPER                 \nINCREASE   DATE / TIME             GALLONS TC GALLONS WATER  TEMP DEG F  HEIGHT\n\n      END: 05/30/2018 14:14         1947       2064   5.32      56.55    65.48\n    START: 05/30/2018 14:04         1347       1464   5.32      56.55    42.480000000000004\n   AMOUNT:                          1647       1764\n\n',
    'I20300': b'\nI20300\n05/30/2018 19:18\n\nSTATOIL STATION\n\n\nTANK 1    SUPER                 \n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\nTANK 2    UNLEAD                \n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\nTANK 3    DIESEL                \n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\nTANK 4    PREMIUM               \n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\n',
    'I20400': b'\nI20400\n05/30/2018 19:18\n\nSTATOIL STATION\n\n\n\nSHIFT REPORT\n\nSHIFT 1 TIME: 12:00 AM\n\nTANK PRODUCT\n\n  1  SUPER                  VOLUME TC VOLUME  ULLAGE  HEIGHT  WATER   TEMP\nSHIFT  1 STARTING VALUES      7950     8130    9372   31.03   1.25    52.60\n         ENDING VALUES        8890     9016    9717   84.03  1.25    52.60\n         DELIVERY VALUE          0\n         TOTALS                940\n\n',
    'I20500': b'\nI20500\n05/30/2018 19:18\n\n\nSTATOIL STATION\n\n\nTANK   PRODUCT                 STATUS\n\n  1    SUPER                   NORMAL\n\n  2    UNLEAD                  HIGH WATER ALARM\n                               HIGH WATER WARNING\n\n  3    DIESEL                  NORMAL\n\n  4    PREMIUM                 NORMAL\n\n'
}


class TestGuardianAST(unittest.TestCase):
    def setUp(self):
        # clean up before we start...
        conpot_core.get_sessionManager().purge_sessions()

        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)
        args = namedtuple('FakeArgs', '')
        self.guardian_ast_server = GuardianASTServer(self.dir_name +
                                                     '/templates/guardian_ast/guardian_ast/guardian_ast.xml',
                                                     self.dir_name + '/templates/guardian_ast', args)
        self.server_greenlet = gevent.spawn(self.guardian_ast_server.start, '127.0.0.1', 0)
        # initialize the databus
        self.guardian_ast_server.databus.initialize(self.dir_name + '/templates/guardian_ast/template.xml')
        gevent.sleep(1)

    def tearDown(self):
        self.guardian_ast_server.stop()
        gevent.joinall([self.server_greenlet])
        # tidy up (again)...
        conpot_core.get_sessionManager().purge_sessions()

    def test_I20100(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01I20100\r\n')
        data = s.recv(1024)
        s.close()
        # FIXME: Omitting the time etc from data - mechanism to check them needed as well?
        self.assertEqual(data[:8] + data[24:156], DATA['I20100'][:8] + DATA['I20100'][24:156])

    def test_I20200(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01I20200\r\n')
        data = s.recv(1024)
        s.close()
        self.assertEqual(data[:8] + data[24:181], DATA['I20200'][:8] + DATA['I20200'][24:181])

    def test_I20300(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01I20300\r\n')
        data = s.recv(1024)
        s.close()
        self.assertEqual(data[:8] + data[24:], DATA['I20300'][:8] + DATA['I20300'][24:])

    def test_I20400(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01I20400\r\n')
        data = s.recv(1024)
        s.close()
        self.assertEqual(data[:8] + data[24:202], DATA['I20400'][:8] + DATA['I20400'][24:202])

    def test_I20500(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01I20500\r\n')
        data = s.recv(1024)
        s.close()
        self.assertEqual(data[:8] + data[24:], DATA['I20500'][:8] + DATA['I20500'][24:])

    def test_ast_error(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01S6020\r\n')
        data = s.recv(1024)
        s.close()
        self.assertEqual(data, b'9999FF1B\n')

    def test_S60201(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.5)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01S60201NONSUPER\r\n')
        try:
            _ = s.recv(1024)
        except socket.timeout:
            pass
        s.send(b'\x01I20100\r\n')
        data = s.recv(1024)
        s.close()
        self.assertIn(b'NONSUPER', data)

    def test_S60202(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.5)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01S60202TESTLEAD\r\n')
        try:
            _ = s.recv(1024)
        except socket.timeout:
            pass
        s.send(b'\x01I20100\r\n')
        data = s.recv(1024)
        s.close()
        self.assertIn(b'TESTLEAD', data)

    def test_S60203(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.5)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01S60203TESTDIESEL\r\n')
        try:
            _ = s.recv(1024)
        except socket.timeout:
            pass
        s.send(b'\x01I20100\r\n')
        data = s.recv(1024)
        s.close()
        self.assertIn(b'TESTDIESEL', data)

    def test_S60204(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.5)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01S60204TESTPREMIUM\r\n')
        try:
            _ = s.recv(1024)
        except socket.timeout:
            pass
        s.send(b'\x01I20100\r\n')
        data = s.recv(1024)
        s.close()
        self.assertIn(b'TESTPREMIUM', data)

    def test_S60200(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.5)
        s.connect(('127.0.0.1', self.guardian_ast_server.server.server_port))
        s.send(b'\x01S60200ULTIMATETEST\r\n')
        try:
            _ = s.recv(1024)
        except socket.timeout:
            pass
        s.send(b'\x01I20100\r\n')
        data = s.recv(1024)
        s.close()
        count = len(re.findall('(?=ULTIMATETEST)', data.decode()))
        self.assertEqual(count, 4)


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2017  Patrick Reichenberger (University of Passau) <patrick.reichenberger@t-online.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


class IEC104Register(object):
    def __init__(self, category_id, addr, val, relation):
        self.category_id = category_id
        self.addr = addr
        self.val = val
        self.relation = relation

    def set_val(self, val):
        self.val = val



# Copyright (C) 2013  Daniel creo Haslinger <creo-conpot@blackmesa.at>
# Derived from plcscan by Dmitry Efanov (Positive Research)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from struct import *
from optparse import OptionGroup
from conpot.helpers import chr_py3, str_to_bytes
import struct
import socket
import string
# from conpot.protocols.s7comm.tpkt import cleanse_byte_string

__FILTER = "".join([' '] + [' ' if chr(x) not in string.printable or chr(x) in string.whitespace else chr(x)
                            for x in range(1, 256)])


def StripUnprintable(msg):
    return msg.translate(__FILTER)


class TPKTPacket:
    """ TPKT packet. RFC 1006
    """
    def __init__(self, data=''):
        self.data = data

    def pack(self):
        return pack('!BBH',
                    3,                         # version
                    0,                         # reserved
                    len(bytes(self.data))+4    # packet size
                    ) + str_to_bytes(bytes(self.data))

    def unpack(self, packet):
        try:
            header = unpack('!BBH', packet[:4])
        except struct.error:
            raise S7ProtocolError("Unknown TPKT format")

        self.data = packet[4:4+header[2]]
        return self


class COTPConnectionPacket:
    """ COTP Connection Request or Connection Confirm packet (ISO on TCP). RFC 1006
    """
    def __init__(self, dst_ref=0, src_ref=0, dst_tsap=0, src_tsap=0, tpdu_size=0):
        self.dst_ref = dst_ref
        self.src_ref = src_ref
        self.dst_tsap = dst_tsap
        self.src_tsap = src_tsap
        self.tpdu_size = tpdu_size

    def pack(self):
        """ make Connection Request Packet
        """
        return pack('!BBHHBBBHBBHBBB',
                    17,             # size
                    0xe0,           # pdu type: CR
                    self.dst_ref,
                    self.src_ref,
                    0,              # flag
                    0xc1, 2, self.src_tsap,
                    0xc2, 2, self.dst_tsap,
                    0xc0, 1, self.tpdu_size)

    def unpack(self, packet):
        """ parse Connection Confirm Packet (header only)
        """
        try:
            size, pdu_type, self.dst_ref, self.src_ref, flags = unpack('!BBHHB', packet[:7])
        except struct.error:
            raise S7ProtocolError("Wrong CC packet format")
        if len(packet) != size + 1:
            raise S7ProtocolError("Wrong CC packet size")
        if pdu_type != 0xd0:
            raise S7ProtocolError("Not a CC packet")
        return self

    def __bytes__(self):
        return self.pack()


class COTPDataPacket:
    """ COTP Data packet (ISO on TCP). RFC 1006
    """
    def __init__(self, data=''):
        self.data = data

    def pack(self):
        return pack('!BBB',
                    2,                      # header len
                    0xf0,                   # data packet
                    0x80) + str_to_bytes(bytes(self.data))

    def unpack(self, packet):
        self.data = packet[packet[0]+1:]
        return self

    def __bytes__(self):
        return self.pack()


class S7Packet:
    """ S7 packet
    """
    def __init__(self, _type=1, req_id=0, parameters='', data=''):
        self.type = _type
        self.req_id = req_id
        self.parameters = parameters
        self.data = data
        self.error = 0

    def pack(self):
        if self.type not in [1, 7]:
            raise S7ProtocolError("Unknown pdu type")
        return (pack('!BBHHHH',
                     0x32,                   # protocol s7 magic
                     self.type,              # pdu-type
                     0,                      # reserved
                     self.req_id,            # request id
                     len(self.parameters),   # parameters length
                     len(self.data)) +       # data length
                str_to_bytes(self.parameters) +
                str_to_bytes(self.data))

    def unpack(self, packet):
        try:
            if packet[1] in [3, 2]:   # pdu-type = response
                header_size = 12
                magic0x32, self.type, reserved, self.req_id, parameters_length, data_length, self.error = \
                    unpack('!BBHHHHH', packet[:header_size])
                if self.error:
                    raise S7Error(self.error)
            elif packet[1] in [1, 7]:
                header_size = 10
                magic0x32, self.type, reserved, self.req_id, parameters_length, data_length = \
                    unpack('!BBHHHH', packet[:header_size])
            else:
                raise S7ProtocolError("Unknown pdu type (%d)" % packet[1])
        except struct.error:
            raise S7ProtocolError("Wrong S7 packet format")

        self.parameters = packet[header_size:header_size+parameters_length]
        self.data = packet[header_size+parameters_length:header_size+parameters_length+data_length]

        return self

    def __bytes__(self):
        return self.pack()


class S7ProtocolError(Exception):
    def __init__(self, message, packet=''):
        self.message = message
        self.packet = packet

    def __str__(self):
        return "[ERROR][S7Protocol] %s" % self.message


class S7Error(Exception):
    _errors = {
        # s7 data errors
        0x05: 'Address Error',
        0x0a: 'Item not available',
        # s7 header errors
        0x8104: 'Context not supported',
        0x8500: 'Wrong PDU size'
    }

    def __init__(self, code):
        self.code = code

    def __str__(self):
        if self.code in S7Error._errors:
            message = S7Error._errors[self.code]
        else:
            message = 'Unknown error'
        return "[ERROR][S7][0x%x] %s" % (self.code, message)


def Split(ar, size):
    """ split sequence into blocks of given size
    """
    return [ar[i:i + size] for i in range(0, len(ar), size)]


class s7:
    def __init__(self, ip, port, src_tsap=0x200, dst_tsap=0x201, timeout=8):
        self.ip = ip
        self.port = port
        self.req_id = 0
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.dst_ref = 0
        self.src_ref = 0x04
        self.dst_tsap = dst_tsap
        self.src_tsap = src_tsap
        self.timeout = timeout

    def Connect(self):
        """ Establish ISO on TCP connection and negotiate PDU
        """
        # sleep(1)
        # self.src_ref = randint(1, 20)
        self.src_ref = 10
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.settimeout(self.timeout)
        self.s.connect((self.ip, self.port))
        self.s.send(TPKTPacket(COTPConnectionPacket(self.dst_ref,
                                                    self.src_ref,
                                                    self.dst_tsap,
                                                    self.src_tsap,
                                                    0x0a)).pack())
        reply = self.s.recv(1024)
        _ = COTPConnectionPacket().unpack(TPKTPacket().unpack(reply).data)

        self.NegotiatePDU()

    def Request(self, _type, parameters='', data=''):
        """ Send s7 request and receive response
        """
        packet = TPKTPacket(COTPDataPacket(S7Packet(_type, self.req_id, parameters, data))).pack()
        self.s.send(packet)
        reply = self.s.recv(1024)
        response = S7Packet().unpack(COTPDataPacket().unpack(TPKTPacket().unpack(reply).data).data)
        if self.req_id != response.req_id:
            raise S7ProtocolError('Sequence ID not correct')
        return response

    def NegotiatePDU(self, pdu=480):
        """ Send negotiate pdu request and receive response. Reply no matter
        """
        response = self.Request(0x01, pack('!BBHHH',
                                           0xf0,       # function NegotiatePDU
                                           0x00,       # unknown
                                           0x01,       # max number of parallel jobs
                                           0x01,       # max number of parallel jobs
                                           pdu))      # pdu length

        func, unknown, pj1, pj2, pdu = unpack('!BBHHH', response.parameters)
        return pdu

    def Function(self, _type, group, function, data=''):
        parameters = pack('!LBBBB',
                          0x00011200 +            # parameter head (magic)
                          0x04,                   # parameter length
                          0x11,                   # unknown
                          _type*0x10+group,        # type, function group
                          function,               # function
                          0x00)                  # sequence

        data = pack('!BBH', 0xFF, 0x09, len(data)) + data
        response = self.Request(0x07, parameters, data)

        code, transport_size, data_len = unpack('!BBH', response.data[:4])
        if code != 0xFF:
            raise S7Error(code)
        return response.data[4:]

    def ReadSZL(self, szl_id):
        szl_data = self.Function(
            0x04,                   # request
            0x04,                   # szl-functions
            0x01,                   # read szl
            pack('!HH',
                 szl_id,             # szl id
                 1))                 # szl index

        szl_id, szl_index, element_size, element_count = unpack('!HHHH', szl_data[:8])

        return Split(szl_data[8:], element_size)


def BruteTsap(ip, port, src_tsaps=(0x100, 0x200), dst_tsaps=(0x102, 0x200, 0x201)):
    for src_tsap in src_tsaps:
        for dst_tsap in dst_tsaps:
            try:
                con = s7(ip, port)
                con.src_tsap = src_tsap
                con.dst_tsap = dst_tsap
                con.Connect()
                return src_tsap, dst_tsap

            except S7ProtocolError:
                pass

    return None


def GetIdentity(ip, port, src_tsap, dst_tsap):
    res = []

    szl_dict = {
        0x11: {
            'title': 'Module Identification',
            'indexes': {
                1: 'Module',
                6: 'Basic Hardware',
                7: 'Basic Firmware'
            },
            'packer': {
                (1, 6): lambda packet: "{0:s} v.{2:d}.{3:d}".format(*unpack('!20sHBBH', packet)),
                (7,): lambda packet: "{0:s} v.{3:d}.{4:d}.{5:d}".format(*unpack('!20sHBBBB', packet))
            }
        },
        0x1c: {
            'title': 'Component Identification',
            'indexes': {
                1: 'Name of the PLC',
                2: 'Name of the module',
                3: 'Plant identification',
                4: 'Copyright',
                5: 'Serial number of module',
                6: 'Reserved for operating system',
                7: 'Module type name',
                8: 'Serial number of memory card',
                9: 'Manufacturer and profile of a CPU module',
                10: 'OEM ID of a module',
                11: 'Location designation of a module'
            },
            'packer': {
                (1, 2, 5): lambda packet: "%s" % packet[:24],
                (3, 7, 8): lambda packet: "%s" % packet[:32],
                (4,): lambda packet: "%s" % packet[:26]
            }
        }
    }

    con = s7(ip, port, src_tsap, dst_tsap)
    con.Connect()

    for szl_id in list(szl_dict.keys()):
        try:
            entities = con.ReadSZL(szl_id)
        except S7Error:
            continue

        indexes = szl_dict[szl_id]['indexes']
        packers = szl_dict[szl_id]['packer']

        for item in entities:
            if len(item) > 2:
                n, = unpack('!H', item[:2])
                item = item[2:]

                try:
                    packers_keys = [i for i in list(packers.keys()) if n in i]
                    formated_item = packers[packers_keys[0]](item).strip('\x00')
                except (struct.error, IndexError):
                    formated_item = StripUnprintable(item).strip('\x00')

                res.append("%s;%s;%s" % (szl_id, n, formated_item))

    return res


def Scan(ip, port):
    res = ()
    try:
        res = BruteTsap(ip, port)
    except socket.error as e:
        print("%s:%d %s" % (ip, port, e))

    if not res:
        print(" MEH!")
        return False

    print("%s:%d S7comm (src_tsap=0x%x, dst_tsap=0x%x)" % (ip, port, res[0], res[1]))

    # sometimes unexpected exceptions occur, so try to get identity several time
    identities = []
    for attempt in [0, 1]:
        try:
            identities = GetIdentity(ip, port, res[0], res[1])
            break
        except (S7ProtocolError, socket.error) as e:
            print("Attempt {0}:  {1}".format(attempt, e))

    return identities


def AddOptions(parser):
    group = OptionGroup(parser, "S7 scanner options")
    group.add_option("--src-tsap", help="Try this src-tsap (list) (default: 0x100,0x200)",
                     type="string", metavar="LIST")
    group.add_option("--dst-tsap", help="Try this dst-tsap (list) (default: 0x102,0x200,0x201)",
                     type="string", metavar="LIST")
    parser.add_option_group(group)



# Command Responder (GET/GETNEXT)
# Based on examples from http://pysnmp.sourceforge.net/

from pysnmp.entity import engine, config
from pysnmp.carrier.asynsock.dgram import udp
from pysnmp.entity.rfc3413 import cmdgen
from pysnmp.proto import rfc1902


class SNMPClient(object):
    def __init__(self, host, port):

        # Create SNMP engine instance
        self.snmpEngine = engine.SnmpEngine()

        # user: usr-sha-aes, auth: SHA, priv AES
        config.addV3User(
            self.snmpEngine, 'usr-sha-aes128',
            config.usmHMACSHAAuthProtocol, 'authkey1',
            config.usmAesCfb128Protocol, 'privkey1'
        )
        config.addTargetParams(self.snmpEngine, 'my-creds', 'usr-sha-aes128', 'authPriv')

        # Setup transport endpoint and bind it with security settings yielding
        # a target name (choose one entry depending of the transport needed).

        # UDP/IPv4
        config.addSocketTransport(
            self.snmpEngine,
            udp.domainName,
            udp.UdpSocketTransport().openClientMode()
        )
        config.addTargetAddr(
            self.snmpEngine, 'my-router',
            udp.domainName, (host, port),
            'my-creds'
        )

    # Error/response receiver
    def cbFun(self, sendRequestHandle, errorIndication, errorStatus, errorIndex, varBindTable, cbCtx):
        if errorIndication:
            print(errorIndication)
        elif errorStatus:
            print(('%s at %s' % (
                errorStatus.prettyPrint(),
                errorIndex and varBindTable[-1][int(errorIndex) - 1] or '?')
            ))
        else:
            for oid, val in varBindTable:
                print(('%s = %s' % (oid.prettyPrint(), val.prettyPrint())))

    def get_command(self, OID=((1, 3, 6, 1, 2, 1, 1, 1, 0), None), callback=None):
        if not callback:
            callback = self.cbFun
            # Prepare and send a request message
        cmdgen.GetCommandGenerator().sendReq(
            self.snmpEngine,
            'my-router',
            (OID,),
            callback,
        )
        self.snmpEngine.transportDispatcher.runDispatcher()
        # Run I/O dispatcher which would send pending queries and process responses
        self.snmpEngine.transportDispatcher.runDispatcher()

    def set_command(self, OID, callback=None):
        if not callback:
            callback = self.cbFun
        cmdgen.SetCommandGenerator().sendReq(
            self.snmpEngine,
            'my-router',
            (OID,),
            callback,
        )
        self.snmpEngine.transportDispatcher.runDispatcher()

    def walk_command(self, OID, callback=None):
        if not callback:
            callback = self.cbFun
        cmdgen.NextCommandGenerator().sendReq(
            self.snmpEngine,
            'my-router',
            (OID,),
            callback,
        )


if __name__ == "__main__":
    snmp_client = SNMPClient('127.0.0.1', 161)
    OID = ((1, 3, 6, 1, 2, 1, 1, 1, 0), None)
    snmp_client.get_command(OID)
